"""Contains ecoMAX parameter descriptors."""

from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING

from dataslots import dataslots

from pyplumio.const import (
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_SIZE,
    ATTR_VALUE,
    PERCENTAGE,
    FrameType,
    ProductType,
    UnitOfMeasurement,
)
from pyplumio.frames import Request
from pyplumio.helpers.async_cache import acache
from pyplumio.parameters import (
    OffsetNumber,
    OffsetNumberDescription,
    Parameter,
    ParameterDescription,
    Switch,
    SwitchDescription,
)
from pyplumio.parameters.custom import inject_custom_parameters
from pyplumio.structures.ecomax_parameters import ATTR_ECOMAX_CONTROL
from pyplumio.structures.product_info import ProductInfo
from pyplumio.structures.thermostat_parameters import ATTR_THERMOSTAT_PROFILE

if TYPE_CHECKING:
    from pyplumio.devices.ecomax import EcoMAX


@dataclass
class EcomaxParameterDescription(ParameterDescription):
    """Represents an ecoMAX parameter description."""

    __slots__ = ()


class EcomaxParameter(Parameter):
    """Represents an ecoMAX parameter."""

    __slots__ = ()

    device: EcoMAX
    description: EcomaxParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        handler = partial(Request.create, recipient=self.device.address)
        if self.description.name == ATTR_ECOMAX_CONTROL:
            return await handler(
                frame_type=FrameType.REQUEST_ECOMAX_CONTROL,
                data={ATTR_VALUE: self.values.value},
            )

        if self.description.name == ATTR_THERMOSTAT_PROFILE:
            return await handler(
                frame_type=FrameType.REQUEST_SET_THERMOSTAT_PARAMETER,
                data={
                    ATTR_INDEX: self._index,
                    ATTR_VALUE: self.values.value,
                    ATTR_OFFSET: 0,
                    ATTR_SIZE: 1,
                },
            )

        return await handler(
            frame_type=FrameType.REQUEST_SET_ECOMAX_PARAMETER,
            data={ATTR_INDEX: self._index, ATTR_VALUE: self.values.value},
        )


@dataclass
class EcomaxNumberDescription(EcomaxParameterDescription, OffsetNumberDescription):
    """Represents an ecoMAX number description."""

    __slots__ = ()


class EcomaxNumber(EcomaxParameter, OffsetNumber):
    """Represents a ecoMAX number."""

    __slots__ = ()

    description: EcomaxNumberDescription


@dataslots
@dataclass
class EcomaxSwitchDescription(EcomaxParameterDescription, SwitchDescription):
    """Represents an ecoMAX switch description."""


class EcomaxSwitch(EcomaxParameter, Switch):
    """Represents an ecoMAX switch."""

    __slots__ = ()

    description: EcomaxSwitchDescription


PARAMETER_TYPES: dict[ProductType, list[EcomaxParameterDescription]] = {
    ProductType.ECOMAX_P: [
        EcomaxNumberDescription(
            name="airflow_power_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_power_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_power_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="boiler_power_100",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="boiler_power_50",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="boiler_power_30",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="max_fan_boiler_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="min_fan_boiler_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_100",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_50",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_30",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_100",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_50",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_30",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="cycle_duration",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="h2_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="h1_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="heating_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="fuzzy_logic",
        ),
        EcomaxNumberDescription(
            name="min_fuzzy_logic_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="max_fuzzy_logic_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="min_boiler_power",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="max_boiler_power",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="min_fan_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="max_fan_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="reduction_airflow_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="fan_power_gain",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuzzy_logic_fuel_flow_correction",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuel_flow_correction",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="grate_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="grate_heating_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="grate_fan_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="grate_fan_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="grate_heating_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="grate_fuel_detection_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="kindling_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="kindling_low_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="kindling_airflow_delay",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="kindling_test_time",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="kindling_feeder_work",
        ),
        EcomaxNumberDescription(
            name="kindling_feeder_dose",
            unit_of_measurement=UnitOfMeasurement.GRAMS,
        ),
        EcomaxNumberDescription(
            name="kindling_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="warming_up_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="kindling_finish_exhaust_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_finish_threshold_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_fumes_delta_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_delta_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_min_power_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="stabilization_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="stabilization_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="supervision_time",
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_dose",
            unit_of_measurement=UnitOfMeasurement.GRAMS,
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="supervision_cycle_duration",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="supervision_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="supervision_fan_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="supervision_fan_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="increase_fan_support_mode",
        ),
        EcomaxNumberDescription(
            name="burning_off_max_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="burning_off_min_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="burning_off_time",
        ),
        EcomaxNumberDescription(
            name="burning_off_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="burning_off_fan_work",
        ),
        EcomaxNumberDescription(
            name="burning_off_fan_pause",
        ),
        EcomaxNumberDescription(
            name="start_burning_off",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="stop_burning_off",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="cleaning_begin_time",
        ),
        EcomaxNumberDescription(
            name="burning_off_cleaning_time",
        ),
        EcomaxNumberDescription(
            name="cleaning_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="warming_up_pause_time",
        ),
        EcomaxNumberDescription(
            name="warming_up_cycle_time",
        ),
        EcomaxNumberDescription(
            name="remind_time",
        ),
        EcomaxSwitchDescription(
            name="lambda_control",
        ),
        EcomaxNumberDescription(
            name="lambda_correction_range",
        ),
        EcomaxNumberDescription(
            name="oxygen_100",
        ),
        EcomaxNumberDescription(
            name="oxygen_50",
        ),
        EcomaxNumberDescription(
            name="oxygen_30",
        ),
        EcomaxNumberDescription(
            name="fuzzy_logic_oxygen_correction",
        ),
        EcomaxNumberDescription(
            name="max_fuel_flow",
            step=0.2,
            unit_of_measurement=UnitOfMeasurement.KILOGRAMS_PER_HOUR,
        ),
        EcomaxNumberDescription(
            name="feeder_calibration",
        ),
        EcomaxNumberDescription(
            name="fuel_factor",
        ),
        EcomaxNumberDescription(
            name="fuel_calorific_value",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.KILO_WATT_HOUR_PER_KILOGRAM,
        ),
        EcomaxNumberDescription(
            name="fuel_detection_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="fuel_detection_exhaust_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="schedule_feeder_2",
        ),
        EcomaxNumberDescription(
            name="feed2_h1",
        ),
        EcomaxNumberDescription(
            name="feed2_h2",
        ),
        EcomaxNumberDescription(
            name="feed2_h3",
        ),
        EcomaxNumberDescription(
            name="feed2_h4",
        ),
        EcomaxNumberDescription(
            name="feed2_work",
        ),
        EcomaxNumberDescription(
            name="feed2_pause",
        ),
        EcomaxNumberDescription(
            name="heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="heating_pump_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="pause_heating_for_water_heater",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="thermostat_pause",
        ),
        EcomaxNumberDescription(
            name="thermostat_work",
        ),
        EcomaxNumberDescription(
            name="increase_heating_temp_for_water_heater",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="weather_control",
        ),
        EcomaxNumberDescription(
            name="heating_curve",
            step=0.1,
        ),
        EcomaxNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="weather_factor",
        ),
        EcomaxNumberDescription(
            name="thermostat_operation",
        ),
        EcomaxSwitchDescription(
            name="thermostat_mode",
        ),
        EcomaxNumberDescription(
            name="thermostat_decrease_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="disable_pump_on_thermostat",
        ),
        EcomaxNumberDescription(
            name="boiler_alert_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_feeder_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="external_boiler_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="alert_notify",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="pump_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_mode",
        ),
        EcomaxNumberDescription(
            name="water_heater_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_disinfection",
        ),
        EcomaxNumberDescription(
            name="summer_mode",
        ),
        EcomaxNumberDescription(
            name="summer_mode_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="summer_mode_disable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_extension",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxSwitchDescription(
            name="circulation_control",
        ),
        EcomaxNumberDescription(
            name="circulation_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_work",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_start_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="buffer_control",
        ),
        EcomaxNumberDescription(
            name="max_buffer_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_buffer_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_load_start",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_load_stop",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ],
    ProductType.ECOMAX_I: [
        EcomaxNumberDescription(
            name="water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_priority",
        ),
        EcomaxNumberDescription(
            name="water_heater_support",
        ),
        EcomaxNumberDescription(
            name="min_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_extension",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="water_heater_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_disinfection",
        ),
        EcomaxNumberDescription(
            name="water_heater_work_mode",
        ),
        EcomaxSwitchDescription(
            name="solar_support",
        ),
        EcomaxNumberDescription(
            name="solar_pump_on_delta_temp",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="solar_pump_off_delta_temp",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_collector_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_collector_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="collector_off_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_pump_revolutions",
        ),
        EcomaxNumberDescription(
            name="solar_antifreeze",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="circulation_control",
        ),
        EcomaxNumberDescription(
            name="circulation_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_work",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_start_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source",
        ),
        EcomaxNumberDescription(
            name="min_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="critical_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source_pump_extension_time",
        ),
        EcomaxNumberDescription(
            name="additional_heat_source",
        ),
        EcomaxNumberDescription(
            name="main_heat_source_off_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="additional_heat_source_pump_startup_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="hydraulic_diagram",
        ),
        EcomaxSwitchDescription(
            name="antifreeze",
        ),
        EcomaxNumberDescription(
            name="antifreeze_delay",
        ),
        EcomaxNumberDescription(
            name="circuit_lock_time",
        ),
        EcomaxNumberDescription(
            name="circuit_work_time",
        ),
        EcomaxNumberDescription(
            name="alert_out_c",
        ),
        EcomaxNumberDescription(
            name="alert_on_out_c",
        ),
        EcomaxNumberDescription(
            name="thermostat_hysteresis",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="critial_additional_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="automatic_pump_lock_time",
        ),
        EcomaxNumberDescription(
            name="summer_mode",
        ),
        EcomaxNumberDescription(
            name="summer_mode_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="summer_mode_disable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ],
}


ECOMAX_CONTROL_PARAMETER = EcomaxSwitchDescription(
    name=ATTR_ECOMAX_CONTROL, optimistic=True
)
THERMOSTAT_PROFILE_PARAMETER = EcomaxNumberDescription(name=ATTR_THERMOSTAT_PROFILE)


@acache
async def get_ecomax_parameter_types(
    product_info: ProductInfo,
) -> list[EcomaxParameterDescription]:
    """Return ecoMAX parameter types for specific product."""
    return await inject_custom_parameters(
        product_info, parameter_types=PARAMETER_TYPES[product_info.type]
    )


__all__ = [
    "ATTR_ECOMAX_CONTROL",
    "EcomaxNumber",
    "EcomaxNumberDescription",
    "EcomaxParameter",
    "EcomaxParameterDescription",
    "EcomaxSwitch",
    "EcomaxSwitchDescription",
    "get_ecomax_parameter_types",
    "PARAMETER_TYPES",
    "THERMOSTAT_PROFILE_PARAMETER",
]



"""Contains tests for the frame reader and writer classes."""

from __future__ import annotations

import asyncio
from collections.abc import Generator
import logging
from unittest.mock import AsyncMock, call, patch

import pytest

from pyplumio.const import DeviceType, FrameType
from pyplumio.exceptions import ChecksumError, ReadError, UnknownDeviceError
from pyplumio.frames import ECONET_TYPE, ECONET_VERSION
from pyplumio.frames.requests import EcomaxParametersRequest, ProgramVersionRequest
from pyplumio.stream import (
    DEFAULT_BUFFER_SIZE,
    MAX_FRAME_LENGTH,
    WRITER_TIMEOUT,
    BufferedReader,
    FrameReader,
    FrameWriter,
)


@pytest.fixture(name="mock_stream_writer")
def fixture_mock_stream_writer() -> Generator[asyncio.StreamWriter, None, None]:
    """Mock asyncio StreamWriter.

    Provides a patched StreamWriter for testing purposes.
    """
    with patch("asyncio.StreamWriter", autospec=True) as mock_stream_writer:
        yield mock_stream_writer


@pytest.fixture(name="frame_writer")
def fixture_frame_writer(mock_stream_writer) -> FrameWriter:
    """FrameWriter instance.

    Returns a FrameWriter using the mocked StreamWriter.
    """
    return FrameWriter(mock_stream_writer)


@pytest.fixture(name="buffered_reader")
def fixture_buffered_reader() -> Generator[BufferedReader]:
    """BufferedReader instance.

    Returns a BufferedReader using a patched StreamReader.
    """
    with patch("asyncio.StreamReader", autospec=True) as mock_stream_reader:
        yield BufferedReader(mock_stream_reader)


@pytest.fixture(name="frame_reader")
def fixture_frame_reader() -> Generator[FrameReader]:
    """FrameReader instance.

    Returns a FrameReader using a patched StreamReader.
    """
    with patch("asyncio.StreamReader", autospec=True) as mock_stream_reader:
        yield FrameReader(mock_stream_reader)


@pytest.fixture(name="read_frame_start")
def fixture_read_frame_start():
    """Mock frame start delimiter.

    Patches StreamReader.read to return the frame start byte.
    """
    with patch("asyncio.StreamReader.read", return_value=b"\x68") as mock_read:
        yield mock_read


class TestFrameWriter:
    """Tests for FrameWriter class.

    Verifies writing, closing, and waiting for close operations.
    """

    async def test_write(self, frame_writer: FrameWriter, mock_stream_writer) -> None:
        """Test writing.

        Ensures FrameWriter writes and drains as expected.
        """
        program_version_request = ProgramVersionRequest()
        await frame_writer.write(program_version_request)
        mock_stream_writer.write.assert_called_once_with(program_version_request.bytes)
        mock_stream_writer.drain.assert_awaited_once()

    @pytest.mark.parametrize(
        ("expect_exception", "exception"),
        [
            (False, None),
            (True, OSError),
            (True, asyncio.TimeoutError),
        ],
    )
    async def test_close(
        self,
        expect_exception: bool,
        exception: type[Exception],
        frame_writer: FrameWriter,
        mock_stream_writer,
        caplog,
    ) -> None:
        """Test closing.

        Checks FrameWriter.close handles exceptions and logs errors.
        """
        if expect_exception:
            mock_stream_writer.close.side_effect = exception

        await frame_writer.close()
        if expect_exception:
            assert (
                "Failed to close the frame writer due to an unexpected error"
                in caplog.text
            )
        else:
            mock_stream_writer.close.assert_called_once()
            mock_stream_writer.wait_closed.assert_awaited_once()

    async def test_wait_closed(
        self, frame_writer: FrameWriter, mock_stream_writer
    ) -> None:
        """Test wait_closed.

        Ensures wait_closed awaits the writer and checks timeout.
        """
        await frame_writer.wait_closed()
        mock_stream_writer.wait_closed.assert_awaited_once()

        # Check for timeout decorator presence and value.
        assert (
            getattr(FrameWriter.wait_closed, "_has_timeout_seconds", None)
            == WRITER_TIMEOUT
        )


class TestBufferedReader:
    """Test for BufferedReader class.

    Verifies handling of reader with an internal buffer.
    """

    @patch("pyplumio.stream.BufferedReader.trim_to")
    @patch("asyncio.StreamReader.readexactly")
    @pytest.mark.parametrize(
        ("size", "data"),
        (
            (0, bytearray()),
            (3, bytearray(b"\x00\x01\x02")),
            (5, bytearray(b"\x00\x01\x02\x03\x04")),
        ),
    )
    async def test_ensure_buffer(
        self,
        mock_readexactly,
        mock_trim_to,
        buffered_reader: BufferedReader,
        size: int,
        data: bytearray,
    ) -> None:
        """Test ensuring buffer size.

        Ensures that the buffer is filled to the requested size and
        that the buffer is trimmed appropriately.
        """
        mock_readexactly.return_value = data
        await buffered_reader.ensure_buffer(size)
        if size == 0:
            mock_readexactly.assert_not_awaited()
            mock_trim_to.assert_not_called()
        else:
            mock_readexactly.assert_awaited_once_with(size)
            mock_trim_to.assert_called_once_with(DEFAULT_BUFFER_SIZE)
            assert len(buffered_reader.buffer) == size

    @patch(
        "asyncio.StreamReader.readexactly",
        side_effect=asyncio.IncompleteReadError(bytearray(), expected=7),
    )
    async def test_ensure_buffer_with_incomplete_read(
        self, mock_readexactly, buffered_reader: BufferedReader
    ) -> None:
        """Test ensuring buffer size with incomplete read.

        Ensures that a ReadError is raised when the read is incomplete.
        """
        with pytest.raises(
            ReadError, match="Incomplete read. Tried to read 5 additional bytes"
        ):
            await buffered_reader.ensure_buffer(5)

        mock_readexactly.assert_awaited_once_with(5)

    @patch("asyncio.StreamReader.readexactly", side_effect=asyncio.CancelledError())
    async def test_ensure_buffer_with_cancelled_error(
        self, mock_readexactly, buffered_reader: BufferedReader, caplog
    ) -> None:
        """Test ensuring buffer size with cancelled error.

        Ensures that CancelledError is raised and logged when the read is cancelled.
        """
        with pytest.raises(asyncio.CancelledError), caplog.at_level(logging.DEBUG):
            await buffered_reader.ensure_buffer(5)

        assert "Read operation cancelled while ensuring buffer" in caplog.text
        mock_readexactly.assert_awaited_once_with(5)

    @patch("asyncio.StreamReader.readexactly", side_effect=OSError())
    async def test_ensure_buffer_with_unexpected_error(
        self, mock_readexactly, buffered_reader: BufferedReader
    ) -> None:
        """Test ensuring buffer size with unexpected error.

        Ensures that OSError is raised when an unexpected error occurs.
        """
        with pytest.raises(
            OSError, match="Serial connection broken while trying to ensure 5 bytes"
        ):
            await buffered_reader.ensure_buffer(5)

        mock_readexactly.assert_awaited_once_with(5)

    async def test_consume(self, buffered_reader: BufferedReader) -> None:
        """Test consuming bytes from the buffer.

        Ensures that bytes are removed from the buffer as expected.
        """
        buffered_reader.buffer.extend(bytearray(b"\x00\x01"))
        assert len(buffered_reader.buffer) == 2
        await buffered_reader.consume(2)
        assert len(buffered_reader.buffer) == 0

    async def test_peek(self, buffered_reader: BufferedReader) -> None:
        """Test peeking bytes from the buffer.

        Ensures that peeking returns the correct bytes without consuming them.
        """
        buffered_reader.buffer.extend(bytearray(b"\x00\x01\x02"))
        peeked_data = await buffered_reader.peek(2)
        assert len(peeked_data) == 2
        assert peeked_data == bytearray(b"\x00\x01")
        assert len(buffered_reader.buffer) == 3

    def test_seek_to(self, buffered_reader: BufferedReader) -> None:
        """Test seeking to a delimiter in the buffer.

        Ensures that the buffer is trimmed up to the delimiter.
        """
        buffered_reader.buffer.extend(bytearray(b"\x00\x01\x02\x03\x04"))
        assert buffered_reader.seek_to(2) is True
        assert len(buffered_reader.buffer) == 3

    def test_trim_to(self, buffered_reader: BufferedReader) -> None:
        """Test trimming the buffer to a specific size.

        Ensures that the buffer is trimmed to the correct length.
        """
        buffered_reader.buffer.extend(bytearray(b"\x00\x01\x02\x03\x04"))
        buffered_reader.trim_to(3)
        assert len(buffered_reader.buffer) == 3
        assert buffered_reader.buffer == bytearray(b"\x02\x03\x04")

    @patch("asyncio.StreamReader.read", return_value=bytearray(b"\x00\x01\x02"))
    @patch("pyplumio.stream.BufferedReader.trim_to")
    async def test_read_into_buffer(
        self, mock_trim_to, mock_read, buffered_reader: BufferedReader
    ) -> None:
        """Test filling the internal buffer with data.

        Ensures that the buffer is filled and trimmed as expected.
        """
        await buffered_reader.read_into_buffer(MAX_FRAME_LENGTH)
        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)
        mock_trim_to.assert_called_once_with(DEFAULT_BUFFER_SIZE)
        assert len(buffered_reader.buffer) == 3

    @patch("asyncio.StreamReader.read", side_effect=asyncio.CancelledError())
    async def test_read_into_buffer_with_cancelled_error(
        self, mock_read, buffered_reader: BufferedReader, caplog
    ) -> None:
        """Test filling the internal buffer with cancelled error.

        Ensures that CancelledError is raised and logged when the read is cancelled.
        """
        with caplog.at_level(logging.DEBUG), pytest.raises(asyncio.CancelledError):
            await buffered_reader.read_into_buffer(MAX_FRAME_LENGTH)

        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)
        assert "Read operation cancelled while filling internal buffer." in caplog.text

    @patch("asyncio.StreamReader.read", side_effect=OSError())
    async def test_read_into_buffer_with_unexpected_error(
        self, mock_read, buffered_reader: BufferedReader, caplog
    ) -> None:
        """Test filling the internal buffer with unexpected error.

        Ensures that OSError is raised when an unexpected error occurs.
        """
        with pytest.raises(
            OSError, match="Serial connection broken while filling internal buffer"
        ):
            await buffered_reader.read_into_buffer(MAX_FRAME_LENGTH)

        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)

    @patch("asyncio.StreamReader.read", return_value=b"")
    async def test_read_into_buffer_with_broken_connection(
        self, mock_read, buffered_reader: BufferedReader, caplog
    ) -> None:
        """Test filling the buffer with broken connection.

        Ensures that OSError is raised and logged when the connection is broken.
        """
        with (
            pytest.raises(OSError, match="Serial connection broken"),
            caplog.at_level(logging.DEBUG),
        ):
            await buffered_reader.read_into_buffer(MAX_FRAME_LENGTH)

        assert "Stream ended while filling internal buffer." in caplog.text
        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)


class TestFrameReader:
    """Tests for FrameReader class.

    Verifies reading, error handling, and frame validation.
    """

    @patch("asyncio.StreamReader.read", return_value=(b"\x00\x68"))
    @patch(
        "asyncio.StreamReader.readexactly",
        side_effect=(b"\x0c\x00\x00\x56\x30\x05", b"\x31\xff\x00\xc9\x16"),
    )
    async def test_read(
        self, mock_readexactly, mock_read, frame_reader: FrameReader
    ) -> None:
        """Test reading a frame.

        Checks correct parsing and field extraction from a frame.
        """
        frame = await frame_reader.read()
        assert isinstance(frame, EcomaxParametersRequest)
        assert frame.frame_type == FrameType.REQUEST_ECOMAX_PARAMETERS
        assert frame.sender == DeviceType.ECONET
        assert frame.econet_type == ECONET_TYPE
        assert frame.recipient == DeviceType.ALL
        assert frame.message == b"\xff\x00"
        assert frame.econet_version == ECONET_VERSION
        assert mock_read.call_count == 1
        mock_read.assert_has_calls([call(MAX_FRAME_LENGTH)])
        assert mock_readexactly.call_count == 2
        mock_readexactly.assert_has_calls([call(6), call(5)])

    @patch("asyncio.StreamReader.read", return_value=False)
    async def test_broken_connection(
        self, mock_read, frame_reader: FrameReader
    ) -> None:
        """Test broken connection.

        Ensures OSError is raised for empty read buffer.
        """
        with pytest.raises(OSError, match="Serial connection broken"):
            await frame_reader.read()

        mock_read.assert_awaited_once()

    @patch(
        "asyncio.StreamReader.readexactly",
        side_effect=asyncio.IncompleteReadError(bytearray(), expected=7),
    )
    async def test_incomplete_header(
        self, mock_readexactly, frame_reader: FrameReader, read_frame_start
    ) -> None:
        """Test incomplete header.

        Ensures ReadError is raised for incomplete frame header.
        """
        with pytest.raises(ReadError, match="Incomplete read"):
            await frame_reader.read()

        read_frame_start.assert_awaited_once()
        mock_readexactly.assert_awaited_once_with(6)

    @patch("asyncio.StreamReader.readexactly", return_value=b"\x03\x00\x00\x56\x30\x05")
    async def test_unexpected_frame_length(
        self, mock_readexactly, frame_reader: FrameReader, read_frame_start
    ) -> None:
        """Test unexpected frame length.

        Ensures ReadError is raised for unexpected frame length.
        """
        with pytest.raises(ReadError, match="Unexpected frame length"):
            await frame_reader.read()

        read_frame_start.assert_awaited_once()
        mock_readexactly.assert_awaited_once_with(6)

    @patch(
        "asyncio.StreamReader.readexactly",
        side_effect=(
            b"\x0c\x00\x00\x56\x30\x05",
            asyncio.IncompleteReadError(bytearray(), 10),
        ),
        new_callable=AsyncMock,
    )
    async def test_incomplete_frame(
        self, mock_readexactly, frame_reader: FrameReader, read_frame_start
    ) -> None:
        """Test incomplete frame.

        Ensures ReadError is raised for incomplete frame data.
        """
        with pytest.raises(ReadError, match="Incomplete read"):
            await frame_reader.read()

        read_frame_start.assert_awaited_once()
        mock_readexactly.assert_has_calls([call(6), call(5)])

    @pytest.mark.parametrize(
        ("target", "expected_exception", "error_pattern"),
        [
            ("asyncio.StreamReader.readexactly", OSError, "Serial connection broken"),
            ("asyncio.StreamReader.read", OSError, "Serial connection broken"),
            ("asyncio.StreamReader.readexactly", asyncio.CancelledError, None),
            ("asyncio.StreamReader.read", asyncio.CancelledError, None),
        ],
    )
    async def test_read_exceptions(
        self,
        frame_reader: FrameReader,
        read_frame_start,
        target: str,
        expected_exception: type[Exception],
        error_pattern: str,
    ) -> None:
        """Test read exceptions."""
        with (
            patch(target, side_effect=expected_exception()),
            pytest.raises(expected_exception, match=error_pattern),
        ):
            await frame_reader.read()

    @patch(
        "asyncio.StreamReader.readexactly",
        side_effect=(b"\x0c\x00\x00\x56\x30\x05", b"\x31\xfe\x00\xc9\x16"),
    )
    async def test_incorrect_checksum(
        self, mock_readexactly, frame_reader: FrameReader, read_frame_start
    ) -> None:
        """Test incorrect checksum.

        Ensures ChecksumError is raised for invalid checksum.
        """
        with pytest.raises(ChecksumError, match="Incorrect frame checksum"):
            await frame_reader.read()

        read_frame_start.assert_awaited_once()
        mock_readexactly.assert_has_calls([call(6), call(5)])

    @patch(
        "asyncio.StreamReader.read",
        return_value=b"\x68\x0a\x00\x10\x56\x30\x05\x01\x10\x16",
    )
    async def test_unknown_recipient(
        self, mock_read, frame_reader: FrameReader
    ) -> None:
        """Test unknown recipient.

        Verifies that the reader returns None when the recipient address
        in the frame is not recognized as a valid device.
        """
        result = await frame_reader.read()
        assert result is None
        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)

    @patch(
        "asyncio.StreamReader.read",
        return_value=b"\x68\x0a\x00\x00\x10\x30\x05\x01\x46\x16",
    )
    async def test_unknown_sender(self, mock_read, frame_reader: FrameReader) -> None:
        """Test unknown sender.

        Verifies that an UnknownDeviceError is raised when the sender
        address in the frame does not match any known device types.
        """
        with pytest.raises(UnknownDeviceError, match="Unknown sender"):
            await frame_reader.read()

        mock_read.assert_awaited_once_with(MAX_FRAME_LENGTH)



"""Contains callback filters."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import suppress
from copy import copy
from decimal import Decimal
import logging
import math
import time
from typing import (
    Any,
    Final,
    Protocol,
    SupportsFloat,
    TypeVar,
    overload,
    runtime_checkable,
)

from typing_extensions import TypeAlias

from pyplumio.helpers.event_manager import Callback
from pyplumio.parameters import Parameter

_LOGGER = logging.getLogger(__name__)

numpy_installed = False
with suppress(ImportError):
    import numpy as np

    _LOGGER.info("Using numpy for improved float precision")
    numpy_installed = True


UNDEFINED: Final = "undefined"


@runtime_checkable
class SupportsSubtraction(Protocol):
    """Supports subtraction operation."""

    __slots__ = ()

    def __sub__(
        self: SupportsSubtraction, other: SupportsSubtraction
    ) -> SupportsSubtraction:
        """Subtract a value."""


@runtime_checkable
class SupportsComparison(Protocol):
    """Supports comparison."""

    __slots__ = ()

    def __hash__(self) -> int:
        """Return a hash of the value."""

    def __eq__(self: SupportsComparison, other: SupportsComparison) -> bool:
        """Compare a value."""


Comparable = TypeVar("Comparable", Parameter, SupportsFloat, SupportsComparison)

DEFAULT_TOLERANCE: Final = 1e-6


@overload
def is_close(old: Parameter, new: Parameter, tolerance: None = None) -> bool: ...


@overload
def is_close(
    old: SupportsFloat, new: SupportsFloat, tolerance: float = DEFAULT_TOLERANCE
) -> bool: ...


@overload
def is_close(
    old: SupportsComparison, new: SupportsComparison, tolerance: None = None
) -> bool: ...


def is_close(
    old: Comparable, new: Comparable, tolerance: float | None = DEFAULT_TOLERANCE
) -> bool:
    """Check if value is significantly changed."""
    if isinstance(old, Parameter) and isinstance(new, Parameter):
        return new.update_pending.is_set() or old.values.__ne__(new.values)

    if tolerance and isinstance(old, SupportsFloat) and isinstance(new, SupportsFloat):
        return not math.isclose(old, new, abs_tol=tolerance)

    return old.__ne__(new)


@overload
def diffence_between(old: list, new: list) -> list: ...


@overload
def diffence_between(
    old: SupportsSubtraction, new: SupportsSubtraction
) -> SupportsSubtraction: ...


def diffence_between(
    old: SupportsSubtraction | list, new: SupportsSubtraction | list
) -> SupportsSubtraction | list | None:
    """Return a difference between values."""
    if isinstance(old, list) and isinstance(new, list):
        return [x for x in new if x not in old]

    if isinstance(old, SupportsSubtraction) and isinstance(new, SupportsSubtraction):
        return new.__sub__(old)

    return None


class Filter(ABC):
    """Represents a filter."""

    __slots__ = ("_callback", "_value")

    _callback: Callback
    _value: Any

    def __init__(self, callback: Callback) -> None:
        """Initialize a new filter."""
        self._callback = callback
        self._value = UNDEFINED

    def __hash__(self) -> int:
        """Return a hash of the filter based on its callback."""
        return hash(self._callback)

    def __eq__(self, other: Any) -> bool:
        """Compare callbacks."""
        if isinstance(other, Filter):
            return self._callback == other._callback

        if callable(other):
            return bool(self._callback == other)

        return NotImplemented

    @abstractmethod
    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""


class _Aggregate(Filter):
    """Represents an aggregate filter.

    Calls a callback with a sum of values collected over a specified
    time period or when sample size limit reached.
    """

    __slots__ = ("_values", "_sample_size", "_timeout", "_last_call_time")

    _values: list[float | int | Decimal]
    _sample_size: int
    _timeout: float
    _last_call_time: float

    def __init__(self, callback: Callback, seconds: float, sample_size: int) -> None:
        """Initialize a new aggregate filter."""
        super().__init__(callback)
        self._last_call_time = time.monotonic()
        self._timeout = seconds
        self._sample_size = sample_size
        self._values = []

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if not isinstance(new_value, (float, int, Decimal)):
            raise TypeError(
                "Aggregate filter can only be used with numeric values, got "
                f"{type(new_value).__name__}: {new_value}"
            )

        current_time = time.monotonic()
        self._values.append(new_value)
        time_since_call = current_time - self._last_call_time
        if time_since_call >= self._timeout or len(self._values) >= self._sample_size:
            result = await self._callback(
                np.sum(self._values) if numpy_installed else sum(self._values)
            )
            self._last_call_time = current_time
            self._values = []
            return result


def aggregate(callback: Callback, seconds: float, sample_size: int) -> _Aggregate:
    """Create a new aggregate filter.

    A callback function will be called with a sum of values collected
    over a specified time period or when sample size limit reached.
    Can only be used with numeric values.

    :param callback: A callback function to be awaited once filter
        conditions are fulfilled
    :type callback: Callback
    :param seconds: A callback will be awaited with a sum of values
        aggregated over this amount of seconds.
    :type seconds: float
    :param sample_size: The maximum number of values to aggregate
        before calling the callback
    :type sample_size: int
    :return: An instance of callable filter
    :rtype: _Aggregate
    """
    return _Aggregate(callback, seconds, sample_size)


class _Clamp(Filter):
    """Represents a clamp filter.

    Calls callback with a value clamped between specified boundaries.
    """

    __slots__ = ("_min_value", "_max_value")

    _min_value: float
    _max_value: float

    def __init__(self, callback: Callback, min_value: float, max_value: float) -> None:
        """Initialize a new Clamp filter."""
        super().__init__(callback)
        self._min_value = min_value
        self._max_value = max_value

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if new_value < self._min_value:
            return await self._callback(self._min_value)

        if new_value > self._max_value:
            return await self._callback(self._max_value)

        return await self._callback(new_value)


def clamp(callback: Callback, min_value: float, max_value: float) -> _Clamp:
    """Create a new clamp filter.

    A callback function will be called and passed value clamped
    between specified boundaries.

    :param callback: A callback function to be awaited on new value
    :type callback: Callback
    :param min_value: A lower boundary
    :type min_value: float
    :param max_value: An upper boundary
    :type max_value: float
    :return: An instance of callable filter
    :rtype: _Clamp
    """
    return _Clamp(callback, min_value, max_value)


_FilterT: TypeAlias = Callable[[Any], bool]


class _Custom(Filter):
    """Represents a custom filter.

    Calls a callback with value, if user-defined filter function
    that's called by this class with the value as an argument
    returns true.
    """

    __slots__ = ("_filter_fn",)

    _filter_fn: _FilterT

    def __init__(self, callback: Callback, filter_fn: _FilterT) -> None:
        """Initialize a new custom filter."""
        super().__init__(callback)
        self._filter_fn = filter_fn

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._filter_fn(new_value):
            await self._callback(new_value)


def custom(callback: Callback, filter_fn: _FilterT) -> _Custom:
    """Create a new custom filter.

    A callback function will be called when a user-defined filter
    function, that's being called with the value as an argument,
    returns true.

    :param callback: A callback function to be awaited when
        filter function return true
    :type callback: Callback
    :param filter_fn: Filter function, that will be called with a
        value and should return `True` to await filter's callback
    :type filter_fn: Callable[[Any], bool]
    :return: An instance of callable filter
    :rtype: _Custom
    """
    return _Custom(callback, filter_fn)


class _Deadband(Filter):
    """Represents a deadband filter.

    Calls a callback only when value is significantly changed from the
    previous callback call.
    """

    __slots__ = ("_tolerance",)

    _tolerance: float

    def __init__(self, callback: Callback, tolerance: float) -> None:
        """Initialize a new value changed filter."""
        self._tolerance = tolerance
        super().__init__(callback)

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if not isinstance(new_value, (float, int, Decimal)):
            raise TypeError(
                "Deadband filter can only be used with numeric values, got "
                f"{type(new_value).__name__}: {new_value}"
            )

        if self._value == UNDEFINED or is_close(
            self._value, new_value, tolerance=self._tolerance
        ):
            self._value = new_value
            return await self._callback(new_value)


def deadband(callback: Callback, tolerance: float) -> _Deadband:
    """Create a new deadband filter.

    A callback function will only be called when the value is significantly changed
    from the previous callback call.

    :param callback: A callback function to be awaited on significant value change
    :type callback: Callback
    :param tolerance: The minimum difference required to trigger the callback
    :type tolerance: float
    :return: An instance of callable filter
    :rtype: _Deadband
    """
    return _Deadband(callback, tolerance)


class _Debounce(Filter):
    """Represents a debounce filter.

    Calls a callback only when value is stabilized across multiple
    filter calls.
    """

    __slots__ = ("_calls", "_min_calls")

    _calls: int
    _min_calls: int

    def __init__(self, callback: Callback, min_calls: int) -> None:
        """Initialize a new debounce filter."""
        super().__init__(callback)
        self._calls = 0
        self._min_calls = min_calls

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or is_close(self._value, new_value):
            self._calls += 1
        else:
            self._calls = 0

        if self._value == UNDEFINED or self._calls >= self._min_calls:
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            self._calls = 0
            return await self._callback(new_value)


def debounce(callback: Callback, min_calls: int) -> _Debounce:
    """Create a new debounce filter.

    A callback function will only be called once the value is stabilized
    across multiple filter calls.

    :param callback: A callback function to be awaited on value change
    :type callback: Callback
    :param min_calls: Value shouldn't change for this amount of
        filter calls
    :type min_calls: int
    :return: An instance of callable filter
    :rtype: _Debounce
    """
    return _Debounce(callback, min_calls)


class _Delta(Filter):
    """Represents a difference filter.

    Calls a callback with a difference between two subsequent values.
    """

    __slots__ = ()

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or is_close(self._value, new_value):
            old_value = self._value
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            if (
                self._value != UNDEFINED
                and (difference := diffence_between(old_value, new_value)) is not None
            ):
                return await self._callback(difference)


def delta(callback: Callback) -> _Delta:
    """Create a new difference filter.

    A callback function will be called with a difference between two
    subsequent values.

    :param callback: A callback function that will be awaited with
        difference between values in two subsequent calls
    :type callback: Callback
    :return: An instance of callable filter
    :rtype: _Delta
    """
    return _Delta(callback)


class _OnChange(Filter):
    """Represents a value changed filter.

    Calls a callback only when value is changed from the
    previous callback call.
    """

    __slots__ = ()

    def __init__(self, callback: Callback) -> None:
        """Initialize a new value changed filter."""
        super().__init__(callback)

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or is_close(self._value, new_value):
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            return await self._callback(new_value)


def on_change(callback: Callback) -> _OnChange:
    """Create a new value changed filter.

    A callback function will only be called if the value is changed from the
    previous call.

    :param callback: A callback function to be awaited on value change
    :type callback: Callback
    :return: An instance of callable filter
    :rtype: _OnChange
    """
    return _OnChange(callback)


class _Throttle(Filter):
    """Represents a throttle filter.

    Calls a callback only when certain amount of seconds passed
    since the last call.
    """

    __slots__ = ("_last_called", "_timeout")

    _last_called: float | None
    _timeout: float

    def __init__(self, callback: Callback, seconds: float) -> None:
        """Initialize a new throttle filter."""
        super().__init__(callback)
        self._last_called = None
        self._timeout = seconds

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        current_timestamp = time.monotonic()
        if (
            self._last_called is None
            or (current_timestamp - self._last_called) >= self._timeout
        ):
            self._last_called = current_timestamp
            return await self._callback(new_value)


def throttle(callback: Callback, seconds: float) -> _Throttle:
    """Create a new throttle filter.

    A callback function will only be called once a certain amount of
    seconds passed since the last call.

    :param callback: A callback function that will be awaited once
        filter conditions are fulfilled
    :type callback: Callback
    :param seconds: A callback will be awaited at most once per
        this amount of seconds
    :type seconds: float
    :return: An instance of callable filter
    :rtype: _Throttle
    """
    return _Throttle(callback, seconds)


__all__ = [
    "Filter",
    "aggregate",
    "clamp",
    "custom",
    "deadband",
    "debounce",
    "delta",
    "on_change",
    "throttle",
]



"""Contains tests for the mixer virtual device."""

import asyncio
import logging
from typing import Final

import pytest

from pyplumio.const import ATTR_SENSORS
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.devices.mixer import Mixer
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.frames.responses import MixerParametersResponse
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.mixer import MixerNumber, MixerParameter, MixerSwitch
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from tests.conftest import class_from_json

MIXER_INDEX: Final = 4


@pytest.fixture(name="mixer")
def fixture_mixer(ecomax: EcoMAX) -> Mixer:
    """Return a Mixer object."""
    return Mixer(asyncio.Queue(), parent=ecomax, index=MIXER_INDEX)


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_mixer_sensors_event_listener(
    mixer: Mixer, sensor_data: SensorDataMessage
) -> None:
    """Test event listener for mixer sensors."""
    mixer_sensors_data = sensor_data.data[ATTR_SENSORS][ATTR_MIXER_SENSORS]
    mixer.dispatch_nowait(ATTR_MIXER_SENSORS, mixer_sensors_data[MIXER_INDEX])
    await mixer.wait_until_done()
    assert mixer.data == {
        "current_temp": 20.0,
        "target_temp": 40,
        "pump": False,
        "mixer_sensors": True,
    }


@class_from_json(
    MixerParametersResponse,
    "responses/mixer_parameters.json",
    arguments=("message",),
)
@pytest.mark.parametrize(
    ("name", "cls", "values"),
    [
        (
            "mixer_target_temp",
            MixerNumber,
            ParameterValues(value=40, min_value=30, max_value=60),
        ),
        (
            "min_target_temp",
            MixerNumber,
            ParameterValues(value=20, min_value=30, max_value=40),
        ),
        (
            "max_target_temp",
            MixerNumber,
            ParameterValues(value=80, min_value=70, max_value=90),
        ),
        (
            "thermostat_decrease_target_temp",
            MixerNumber,
            ParameterValues(value=20, min_value=10, max_value=30),
        ),
        (
            "weather_control",
            MixerSwitch,
            ParameterValues(value=1, min_value=0, max_value=1),
        ),
        (
            "heating_curve",
            MixerNumber,
            ParameterValues(value=13, min_value=10, max_value=30),
        ),
    ],
)
async def test_mixer_parameters_event_listener(
    mixer: Mixer,
    mixer_parameters: MixerParametersResponse,
    name: str,
    cls: type[MixerParameter],
    values: ParameterValues,
) -> None:
    """Test event listener for mixer parameters."""
    mixer_parameters_data = mixer_parameters.data[ATTR_MIXER_PARAMETERS]
    mixer.dispatch_nowait(ATTR_MIXER_PARAMETERS, mixer_parameters_data[0])
    await mixer.wait_until_done()
    assert len(mixer.data) == 7
    parameter = mixer.get_nowait(name)
    assert isinstance(parameter, cls)
    assert parameter.values == values


@class_from_json(
    MixerParametersResponse,
    "unknown/unknown_mixer_parameter.json",
    arguments=("message",),
)
async def test_mixer_parameters_event_listener_unknown_parameter(
    mixer: Mixer, unknown_mixer_parameter: MixerParametersResponse, caplog
) -> None:
    """Test unknown mixer parameter."""
    mixer_parameters_data = unknown_mixer_parameter.data[ATTR_MIXER_PARAMETERS]
    with caplog.at_level(logging.WARNING):
        mixer.dispatch_nowait(ATTR_MIXER_PARAMETERS, mixer_parameters_data[0])
        await mixer.wait_until_done()

    assert "unknown mixer parameter (14)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text



"""Contains various utility methods."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Mapping
from functools import wraps
from typing import TypeVar

from typing_extensions import ParamSpec

KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


def ensure_dict(initial: dict[KT, VT] | None, *args: dict[KT, VT]) -> dict[KT, VT]:
    """Create or merge multiple dictionaries."""
    data = initial if initial is not None else {}
    for extra in args:
        data |= extra

    return data


def to_camelcase(text: str, overrides: Mapping[str, str] | None = None) -> str:
    """Convert snake_case to CamelCase."""
    if not overrides:
        return "".join((x.capitalize() or "_") for x in text.split("_"))

    return "".join(
        (x.capitalize() or "_") if x.lower() not in overrides else overrides[x.lower()]
        for x in text.split("_")
    )


def is_divisible(a: float, b: float, precision: int = 6) -> bool:
    """Check if a is divisible by b."""
    scale: int = 10**precision
    b_scaled = round(b * scale)
    if b_scaled == 0:
        raise ValueError("Division by zero is not allowed.")

    a_scaled = round(a * scale)
    return a_scaled % b_scaled == 0


T = TypeVar("T")
P = ParamSpec("P")


def timeout(
    seconds: float,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorate a timeout for the awaitable."""

    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await asyncio.wait_for(func(*args, **kwargs), timeout=seconds)

        setattr(wrapper, "_has_timeout_seconds", seconds)
        return wrapper

    return decorator


__all__ = ["ensure_dict", "is_divisible", "to_camelcase", "timeout"]



"""Contains tests for the UID helper class."""

import pytest

from pyplumio.structures.product_info import unpack_uid


@pytest.mark.parametrize(
    ("message", "uid"),
    [
        (bytearray.fromhex("001600110D383338365539"), "D251PAKR3GCPZ1K8G05G0"),
        (bytearray.fromhex("002500300E191932135831"), "CE71HB09J468P1ZZ00980"),
    ],
)
def test_from_bytes(message: bytearray, uid: str) -> None:
    """Test unpacking an UID from bytes."""
    assert unpack_uid(message) == uid



"""Contains test for fan power structure decoder."""

import pytest

from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.fan_power import ATTR_FAN_POWER, FanPowerStructure


@pytest.fixture(name="fan_power_structure")
def fixture_fan_power_structure() -> FanPowerStructure:
    """Fixture for FanPowerStructure."""
    return FanPowerStructure(frame=SensorDataMessage())


class TestFanPowerStructure:
    """Test the FanPower structure decoder."""

    def test_decode(self, fan_power_structure: FanPowerStructure) -> None:
        """Test decoding of fan power structure."""
        message = bytearray([0x00, 0x00, 0x48, 0x42])
        data, offset = fan_power_structure.decode(message)
        assert data == {ATTR_FAN_POWER: 50.0}
        assert offset == 4

    def test_decode_with_nan(self, fan_power_structure: FanPowerStructure) -> None:
        """Test decoding of fan power structure with NaN value."""
        message = bytearray([0x00, 0x00, 0xC0, 0x7F])  # Represents NaN
        data, offset = fan_power_structure.decode(message)
        assert data == {}
        assert offset == 4



"""Contains a regulator data structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.data_types import BitArray, DataType
from pyplumio.structures import StructureDecoder
from pyplumio.structures.frame_versions import FrameVersionsStructure
from pyplumio.structures.regulator_data_schema import ATTR_REGDATA_SCHEMA
from pyplumio.utils import ensure_dict

ATTR_REGDATA: Final = "regdata"

REGDATA_VERSION: Final = "1.0"


class RegulatorDataStructure(StructureDecoder):
    """Represents a regulator data structure."""

    __slots__ = ("_offset", "_bitarray_index")

    _offset: int
    _bitarray_index: int

    def _unpack_regulator_data(self, message: bytearray, data_type: DataType) -> Any:
        """Unpack a regulator data sensor."""
        if not isinstance(data_type, BitArray) and self._bitarray_index > 0:
            # Current data type is not bitarray, but previous was, thus
            # we skip a single byte that was left from the bitarray
            # and reset bitarray index.
            self._offset += 1
            self._bitarray_index = 0

        data_type.unpack(message[self._offset :])
        if isinstance(data_type, BitArray):
            self._bitarray_index = data_type.next(self._bitarray_index)

        self._offset += data_type.size
        return data_type.value

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        data = ensure_dict(data)
        offset += 2
        regdata_version = f"{message[offset + 1]}.{message[offset]}"
        if regdata_version != REGDATA_VERSION:
            return data, offset

        data, self._offset = FrameVersionsStructure(self.frame).decode(
            message, offset + 2, data
        )

        if (device := self.frame.handler) is not None and (
            schema := device.get_nowait(ATTR_REGDATA_SCHEMA, [])
        ):
            self._bitarray_index = 0
            data[ATTR_REGDATA] = {
                param_id: self._unpack_regulator_data(message, data_type)
                for param_id, data_type in schema
            }

        return data, self._offset


__all__ = ["ATTR_REGDATA", "REGDATA_VERSION", "RegulatorDataStructure"]



"""Contains a tests for the request frame classes."""

import pytest
from tests.conftest import load_json_parameters

from pyplumio.const import DeviceType
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import Request
from pyplumio.frames.requests import (
    AlertsRequest,
    CheckDeviceRequest,
    EcomaxControlRequest,
    EcomaxParametersRequest,
    MixerParametersRequest,
    ProgramVersionRequest,
    SetEcomaxParameterRequest,
    SetMixerParameterRequest,
    SetScheduleRequest,
    SetThermostatParameterRequest,
    ThermostatParametersRequest,
)
from pyplumio.frames.responses import DeviceAvailableResponse, ProgramVersionResponse


def test_request_class_response_property() -> None:
    """Test response property for a abstract request class."""
    assert Request().response() is None


def test_program_version_response_recipient_and_type() -> None:
    """Test if program version response recipient and type is set."""
    frame = ProgramVersionRequest(recipient=DeviceType.ALL, sender=DeviceType.ECONET)
    response = frame.response()
    assert isinstance(response, ProgramVersionResponse)
    assert response.recipient == DeviceType.ECONET


def test_check_device_response_recipient_and_type() -> None:
    """Test if check device response recipient and type is set."""
    frame = CheckDeviceRequest(recipient=DeviceType.ALL, sender=DeviceType.ECONET)
    response = frame.response()
    assert isinstance(response, DeviceAvailableResponse)
    assert response.recipient == DeviceType.ECONET


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/ecomax_control.json"),
)
def test_ecomax_control(message, data) -> None:
    """Test an ecoMAX control parameter request."""
    assert EcomaxControlRequest(data=data).message == message


def test_ecomax_control_without_data() -> None:
    """Test an ecoMAX control request without any data."""
    with pytest.raises(FrameDataError):
        getattr(EcomaxControlRequest(), "message")


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/ecomax_parameters.json"),
)
def test_ecomax_parameters(message, data) -> None:
    """Test an ecoMAX parameters request."""
    assert EcomaxParametersRequest(data=data).message == message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/set_ecomax_parameter.json"),
)
def test_set_ecomax_parameter(message, data) -> None:
    """Test a set ecoMAX parameter request."""
    assert SetEcomaxParameterRequest(data=data).message == message


def test_set_ecomax_parameter_without_data() -> None:
    """Test a set ecoMAX parameter request without any data."""
    with pytest.raises(FrameDataError):
        getattr(SetEcomaxParameterRequest(), "message")


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/mixer_parameters.json"),
)
def test_mixer_parameters_request(message, data) -> None:
    """Test a mixer parameters request."""
    assert MixerParametersRequest(data=data).message == message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/set_mixer_parameter.json"),
)
def test_set_mixer_parameter(message, data) -> None:
    """Test a set mixer parameter request."""
    assert SetMixerParameterRequest(data=data).message == message


def test_set_mixer_parameter_without_data() -> None:
    """Test a set mixer parameter request without any data."""
    with pytest.raises(FrameDataError):
        getattr(SetMixerParameterRequest(), "message")


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/set_schedule.json"),
)
def test_set_schedule(message, data) -> None:
    """Test a set schedule request bytes."""
    assert SetScheduleRequest(data=data).message == message


def test_set_schedule_without_data() -> None:
    """Test a set schedule request without any data."""
    with pytest.raises(FrameDataError):
        getattr(SetScheduleRequest(), "message")


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/thermostat_parameters.json"),
)
def test_thermostat_parameters_request(message, data) -> None:
    """Test a thermostat parameters request."""
    assert ThermostatParametersRequest(data=data).message == message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/set_thermostat_parameter.json"),
)
def test_set_thermostat_parameter(message, data) -> None:
    """Test a set thermostat parameter request."""
    assert SetThermostatParameterRequest(data=data).message == message


def test_set_thermostat_parameter_without_data() -> None:
    """Test a set thermostat parameter request without any data."""
    with pytest.raises(FrameDataError):
        getattr(SetThermostatParameterRequest(), "message")


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("requests/alerts.json"),
)
def test_alerts_request(message, data) -> None:
    """Test an alerts request."""
    assert AlertsRequest(data=data).message == message



"""Contains an ecoMAX class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator, Iterable
import logging
import time
from typing import Any, Final

from pyplumio.const import (
    ATTR_FRAME_ERRORS,
    ATTR_PASSWORD,
    ATTR_SENSORS,
    ATTR_SETUP,
    STATE_OFF,
    STATE_ON,
    DeviceState,
    DeviceType,
    FrameType,
    State,
)
from pyplumio.devices import PhysicalDevice
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.exceptions import RequestError
from pyplumio.filters import on_change
from pyplumio.frames import DataFrameDescription, Frame, Request
from pyplumio.helpers.event_manager import event_listener
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.ecomax import (
    ECOMAX_CONTROL_PARAMETER,
    THERMOSTAT_PROFILE_PARAMETER,
    EcomaxNumber,
    EcomaxSwitch,
    EcomaxSwitchDescription,
    get_ecomax_parameter_types,
)
from pyplumio.structures.alerts import ATTR_TOTAL_ALERTS
from pyplumio.structures.ecomax_parameters import (
    ATTR_ECOMAX_CONTROL,
    ATTR_ECOMAX_PARAMETERS,
)
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.network_info import ATTR_NETWORK, NetworkInfo
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo
from pyplumio.structures.regulator_data_schema import ATTR_REGDATA_SCHEMA
from pyplumio.structures.schedules import (
    ATTR_SCHEDULES,
    SCHEDULE_PARAMETERS,
    SCHEDULES,
    Schedule,
    ScheduleDay,
    ScheduleNumber,
    ScheduleSwitch,
    ScheduleSwitchDescription,
)
from pyplumio.structures.thermostat_parameters import ATTR_THERMOSTAT_PARAMETERS
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTAT_SENSORS

_LOGGER = logging.getLogger(__name__)


ATTR_MIXERS: Final = "mixers"
ATTR_THERMOSTATS: Final = "thermostats"
ATTR_FUEL_BURNED: Final = "fuel_burned"

MAX_TIME_SINCE_LAST_FUEL_UPDATE: Final = 5 * 60


class FuelMeter:
    """Represents a fuel meter.

    Calculates the fuel burned based on the time
    elapsed since the last sensor message, which contains fuel
    consumption data. If the elapsed time is within the acceptable
    range, it returns the fuel burned data. Otherwise, it logs a
    warning and returns None.
    """

    __slots__ = ("_last_update_time",)

    _last_update_time: float

    def __init__(self) -> None:
        """Initialize a new fuel meter."""
        self._last_update_time = time.monotonic()

    def calculate(self, fuel_consumption: float) -> float | None:
        """Calculate the amount of burned fuel since last update."""
        current_time = time.monotonic()
        time_since_update = current_time - self._last_update_time
        self._last_update_time = current_time
        if time_since_update < MAX_TIME_SINCE_LAST_FUEL_UPDATE:
            return fuel_consumption * (time_since_update / 3600)

        _LOGGER.warning(
            "Skipping outdated fuel consumption data (was %f seconds old)",
            time_since_update,
        )
        return None


REQUIRED: tuple[DataFrameDescription, ...] = (
    DataFrameDescription(
        frame_type=FrameType.REQUEST_UID,
        provides=ATTR_PRODUCT,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_REGULATOR_DATA_SCHEMA,
        provides=ATTR_REGDATA_SCHEMA,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ECOMAX_PARAMETERS,
        provides=ATTR_ECOMAX_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ALERTS,
        provides=ATTR_TOTAL_ALERTS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_SCHEDULES,
        provides=ATTR_SCHEDULES,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_MIXER_PARAMETERS,
        provides=ATTR_MIXER_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_THERMOSTAT_PARAMETERS,
        provides=ATTR_THERMOSTAT_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_PASSWORD,
        provides=ATTR_PASSWORD,
    ),
)

REQUIRED_TYPES = [description.frame_type for description in REQUIRED]


class EcoMAX(PhysicalDevice):
    """Represents an ecoMAX controller."""

    __slots__ = ("_fuel_meter",)

    _fuel_meter: FuelMeter

    address = DeviceType.ECOMAX

    def __init__(self, queue: asyncio.Queue[Frame], network: NetworkInfo) -> None:
        """Initialize a new ecoMAX controller."""
        super().__init__(queue, network)
        self._fuel_meter = FuelMeter()

    def handle_frame(self, frame: Frame) -> None:
        """Handle frame received from the ecoMAX device."""
        if isinstance(frame, Request) and (
            response := frame.response(data={ATTR_NETWORK: self._network})
        ):
            self.queue.put_nowait(response)

        super().handle_frame(frame)

    def _mixers(self, indexes: Iterable[int]) -> Generator[Mixer, None, None]:
        """Iterate through the mixer indexes.

        For each index, return or create an instance of the mixer class.
        Once done, dispatch the 'mixers' event without waiting.
        """
        mixers: dict[int, Mixer] = self.data.setdefault(ATTR_MIXERS, {})
        for index in indexes:
            yield mixers.setdefault(index, Mixer(self.queue, parent=self, index=index))

        return self.dispatch_nowait(ATTR_MIXERS, mixers)

    def _thermostats(self, indexes: Iterable[int]) -> Generator[Thermostat, None, None]:
        """Iterate through the thermostat indexes.

        For each index, return or create an instance of the thermostat
        class. Once done, dispatch the 'thermostats' event without
        waiting.
        """
        thermostats: dict[int, Thermostat] = self.data.setdefault(ATTR_THERMOSTATS, {})
        for index in indexes:
            yield thermostats.setdefault(
                index, Thermostat(self.queue, parent=self, index=index)
            )

        return self.dispatch_nowait(ATTR_THERMOSTATS, thermostats)

    async def _request_frame_version(
        self, frame_type: FrameType | int, version: int
    ) -> None:
        """Request frame version from the device."""
        setup_done = self.get_nowait(ATTR_SETUP, False)
        if setup_done or frame_type not in REQUIRED_TYPES:
            await super()._request_frame_version(frame_type, version)

    async def _set_ecomax_state(self, state: State) -> bool:
        """Try to set the ecoMAX control state."""
        try:
            switch: EcomaxSwitch = self.data[ATTR_ECOMAX_CONTROL]
            return await switch.set(state)
        except KeyError:
            _LOGGER.error("ecoMAX control is not available. Please try again later.")

        return False

    async def turn_on(self) -> bool:
        """Turn on the ecoMAX controller."""
        return await self._set_ecomax_state(STATE_ON)

    async def turn_off(self) -> bool:
        """Turn off the ecoMAX controller."""
        return await self._set_ecomax_state(STATE_OFF)

    def turn_on_nowait(self) -> None:
        """Turn on the ecoMAX controller without waiting."""
        self.create_task(self.turn_on())

    def turn_off_nowait(self) -> None:
        """Turn off the ecoMAX controller without waiting."""
        self.create_task(self.turn_off())

    async def shutdown(self) -> None:
        """Shutdown tasks for the ecoMAX controller and sub-devices."""
        mixers: dict[str, Mixer] = self.get_nowait(ATTR_MIXERS, {})
        thermostats: dict[str, Thermostat] = self.get_nowait(ATTR_THERMOSTATS, {})
        devices = (mixers | thermostats).values()
        await asyncio.gather(*(device.shutdown() for device in devices))
        await super().shutdown()

    @event_listener
    async def on_event_setup(self, setup: bool) -> None:
        """Request frames required to set up an ecoMAX entry."""
        _LOGGER.debug("Setting up device entry")
        await self.wait_for(ATTR_SENSORS)
        results = await asyncio.gather(
            *(
                self.request(description.provides, description.frame_type)
                for description in REQUIRED
            ),
            return_exceptions=True,
        )

        errors = [
            result.frame_type for result in results if isinstance(result, RequestError)
        ]

        if errors:
            self.dispatch_nowait(ATTR_FRAME_ERRORS, errors)

        _LOGGER.debug("Device entry setup done")

    @event_listener
    async def on_event_ecomax_parameters(
        self, parameters: list[tuple[int, ParameterValues]]
    ) -> bool:
        """Update ecoMAX parameters and dispatch the events."""
        _LOGGER.debug("Received device parameters")
        product_info: ProductInfo = await self.get(ATTR_PRODUCT)
        parameter_types = await get_ecomax_parameter_types(product_info)

        def _ecomax_parameter_events() -> Generator[Coroutine]:
            """Get dispatch calls for ecoMAX parameter events."""
            for index, values in parameters:
                try:
                    description = parameter_types[index]
                except IndexError:
                    _LOGGER.warning(
                        "Encountered unknown ecoMAX parameter (%i): %s. "
                        "Your device isn't fully compatible with this software "
                        "and may not work properly. Please visit the issue tracker "
                        "and open a feature request to support %s",
                        index,
                        values,
                        product_info.model,
                    )
                    return

                handler = (
                    EcomaxSwitch
                    if isinstance(description, EcomaxSwitchDescription)
                    else EcomaxNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self, description=description, values=values, index=index
                    ),
                )

        await asyncio.gather(*_ecomax_parameter_events())
        return True

    @event_listener
    async def on_event_fuel_consumption(self, fuel_consumption: float) -> None:
        """Update the amount of burned fuel."""
        fuel_burned = self._fuel_meter.calculate(fuel_consumption)
        if fuel_burned is not None:
            self.dispatch_nowait(ATTR_FUEL_BURNED, fuel_burned)

    @event_listener
    async def on_event_mixer_parameters(
        self, parameters: dict[int, Any] | None
    ) -> bool:
        """Handle mixer parameters and dispatch the events."""
        _LOGGER.debug("Received mixer parameters")
        if parameters:
            await asyncio.gather(
                *(
                    mixer.dispatch(ATTR_MIXER_PARAMETERS, parameters[mixer.index])
                    for mixer in self._mixers(indexes=parameters.keys())
                )
            )
            return True

        return False

    @event_listener
    async def on_event_mixer_sensors(self, sensors: dict[int, Any] | None) -> bool:
        """Update mixer sensors and dispatch the events."""
        _LOGGER.debug("Received mixer sensors")
        if sensors:
            await asyncio.gather(
                *(
                    mixer.dispatch(ATTR_MIXER_SENSORS, sensors[mixer.index])
                    for mixer in self._mixers(indexes=sensors.keys())
                )
            )
            return True

        return False

    @event_listener
    async def on_event_schedule_parameters(
        self, parameters: list[tuple[int, ParameterValues]]
    ) -> bool:
        """Update schedule parameters and dispatch the events."""

        def _schedule_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for schedule parameter events."""
            for index, values in parameters:
                description = SCHEDULE_PARAMETERS[index]
                handler = (
                    ScheduleSwitch
                    if isinstance(description, ScheduleSwitchDescription)
                    else ScheduleNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self, description=description, values=values, index=index
                    ),
                )

        await asyncio.gather(*_schedule_parameter_events())
        return True

    @event_listener
    async def on_event_sensors(self, sensors: dict[str, Any]) -> bool:
        """Update ecoMAX sensors and dispatch the events."""
        _LOGGER.debug("Received device sensors")
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    @event_listener
    async def on_event_thermostat_parameters(
        self, parameters: dict[int, Any] | None
    ) -> bool:
        """Handle thermostat parameters and dispatch the events."""
        _LOGGER.debug("Received thermostat parameters")
        if parameters:
            await asyncio.gather(
                *(
                    thermostat.dispatch(
                        ATTR_THERMOSTAT_PARAMETERS, parameters[thermostat.index]
                    )
                    for thermostat in self._thermostats(indexes=parameters.keys())
                )
            )
            return True

        return False

    @event_listener
    async def on_event_thermostat_profile(
        self, values: ParameterValues | None
    ) -> EcomaxNumber | None:
        """Update thermostat profile parameter."""
        if values:
            return EcomaxNumber(
                device=self, description=THERMOSTAT_PROFILE_PARAMETER, values=values
            )

        return None

    @event_listener
    async def on_event_thermostat_sensors(self, sensors: dict[int, Any] | None) -> bool:
        """Update thermostat sensors and dispatch the events."""
        _LOGGER.debug("Received thermostat sensors")
        if sensors:
            await asyncio.gather(
                *(
                    thermostat.dispatch(
                        ATTR_THERMOSTAT_SENSORS, sensors[thermostat.index]
                    )
                    for thermostat in self._thermostats(indexes=sensors.keys())
                ),
                return_exceptions=True,
            )
            return True

        return False

    @event_listener
    async def on_event_schedules(
        self, schedules: list[tuple[int, list[list[bool]]]]
    ) -> dict[str, Schedule]:
        """Update schedules."""
        _LOGGER.debug("Received device schedules")
        return {
            SCHEDULES[index]: Schedule(
                name=SCHEDULES[index],
                device=self,
                monday=ScheduleDay.from_iterable(schedule[1]),
                tuesday=ScheduleDay.from_iterable(schedule[2]),
                wednesday=ScheduleDay.from_iterable(schedule[3]),
                thursday=ScheduleDay.from_iterable(schedule[4]),
                friday=ScheduleDay.from_iterable(schedule[5]),
                saturday=ScheduleDay.from_iterable(schedule[6]),
                sunday=ScheduleDay.from_iterable(schedule[0]),
            )
            for index, schedule in schedules
        }

    @event_listener(filter=on_change)
    async def on_event_state(self, state: DeviceState) -> None:
        """Update the ecoMAX control parameter."""
        await self.dispatch(
            ECOMAX_CONTROL_PARAMETER.name,
            EcomaxSwitch.create_or_update(
                description=ECOMAX_CONTROL_PARAMETER,
                device=self,
                values=ParameterValues(
                    value=int(state != DeviceState.OFF), min_value=0, max_value=1
                ),
            ),
        )


__all__ = ["ATTR_MIXERS", "ATTR_THERMOSTATS", "ATTR_FUEL_BURNED", "EcoMAX"]



"""Custom parameters for products."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
import logging
from typing import ClassVar, TypeVar, cast

from pyplumio.helpers.factory import create_instance
from pyplumio.parameters import ParameterDescription
from pyplumio.structures.product_info import ProductInfo
from pyplumio.utils import to_camelcase

_LOGGER = logging.getLogger(__name__)


@dataclass
class Signature:
    """Represents a product signature."""

    __slots__ = ("id", "model")

    id: int
    model: str


@dataclass
class CustomParameter:
    """Represents a custom parameter."""

    __slots__ = ("original", "replacement")

    original: str
    replacement: ParameterDescription


class CustomParameters:
    """Represents a custom parameters."""

    signature: ClassVar[Signature]
    replacements: ClassVar[Sequence[CustomParameter]]

    def validate(self, product_info: ProductInfo) -> bool:
        """Validate the product info."""
        return (
            self.signature.id == product_info.id
            and self.signature.model == product_info.model
        )


async def _load_custom_parameters(
    product_info: ProductInfo,
) -> dict[str, ParameterDescription] | None:
    """Load custom parameters."""
    module_name = product_info.model.replace("-", "_").replace(" ", "_").lower()
    module_path = f"parameters.custom.{module_name}"
    class_name = to_camelcase(module_name).upper().replace("ECOMAX", "EcoMAX")
    class_path = f"{module_path}.{class_name}"
    try:
        _LOGGER.debug(
            "Trying to load custom parameters for %s from %s",
            product_info.model,
            class_path,
        )
        custom_parameters = await create_instance(class_path, cls=CustomParameters)
        if not custom_parameters.validate(product_info):
            raise ValueError
    except (ImportError, TypeError, ValueError):
        _LOGGER.debug("No custom parameters found for %s", product_info.model)
        return None

    return {
        custom_parameter.original: custom_parameter.replacement
        for custom_parameter in custom_parameters.replacements
    }


_DescriptionT = TypeVar("_DescriptionT", bound=ParameterDescription)


async def inject_custom_parameters(
    product_info: ProductInfo, parameter_types: list[_DescriptionT]
) -> list[_DescriptionT]:
    """Patch the parameter types based on the provided overrides."""
    if custom_parameters := await _load_custom_parameters(product_info):
        _LOGGER.debug("Custom parameters found for %s", product_info.model)
        return cast(
            list[_DescriptionT],
            [
                replacement
                if original.name in custom_parameters
                and (replacement := custom_parameters[original.name])
                and (base_class := original.__class__.__bases__[0])
                and isinstance(replacement, base_class)
                else original
                for original in parameter_types
            ],
        )

    return parameter_types


__all__ = (
    "inject_custom_parameters",
    "CustomParameters",
    "CustomParameter",
    "Signature",
)



"""Contains a thermostat parameters structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from typing import Any, Final

from pyplumio.parameters import ParameterValues, unpack_parameter
from pyplumio.parameters.thermostat import get_thermostat_parameter_types
from pyplumio.structures import StructureDecoder
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTATS_AVAILABLE
from pyplumio.utils import ensure_dict

ATTR_THERMOSTAT_PROFILE: Final = "thermostat_profile"
ATTR_THERMOSTAT_PARAMETERS: Final = "thermostat_parameters"

THERMOSTAT_PARAMETER_SIZE: Final = 3


class ThermostatParametersStructure(StructureDecoder):
    """Represents a thermostat parameters data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _thermostat_parameter(
        self, message: bytearray, thermostats: int, start: int, end: int
    ) -> Generator[tuple[int, ParameterValues], None, None]:
        """Get a single thermostat parameter."""
        parameter_types = get_thermostat_parameter_types()
        for index in range(start, (start + end) // thermostats):
            description = parameter_types[index]
            if parameter := unpack_parameter(
                message, self._offset, size=description.size
            ):
                yield (index, parameter)

            self._offset += THERMOSTAT_PARAMETER_SIZE * description.size

    def _thermostat_parameters(
        self, message: bytearray, thermostats: int, start: int, end: int
    ) -> Generator[tuple[int, list[tuple[int, ParameterValues]]], None, None]:
        """Get parameters for a thermostat."""
        for index in range(thermostats):
            if parameters := list(
                self._thermostat_parameter(message, thermostats, start, end)
            ):
                yield (index, parameters)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        device = self.frame.handler
        if not device or not (
            thermostats := device.get_nowait(ATTR_THERMOSTATS_AVAILABLE, None)
        ):
            return ensure_dict(data, {ATTR_THERMOSTAT_PARAMETERS: None}), offset

        start = message[offset + 1]
        end = message[offset + 2]
        offset += 3
        thermostat_profile = unpack_parameter(message, offset)
        self._offset = offset + THERMOSTAT_PARAMETER_SIZE
        return (
            ensure_dict(
                data,
                {
                    ATTR_THERMOSTAT_PROFILE: thermostat_profile,
                    ATTR_THERMOSTAT_PARAMETERS: dict(
                        self._thermostat_parameters(message, thermostats, start, end)
                    ),
                },
            ),
            self._offset,
        )


__all__ = [
    "ATTR_THERMOSTAT_PROFILE",
    "ATTR_THERMOSTAT_PARAMETERS",
    "ThermostatParametersStructure",
]



"""Contains a thermostat class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator
import logging
from typing import Any

from pyplumio.devices import VirtualDevice
from pyplumio.helpers.event_manager import event_listener
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.thermostat import (
    ThermostatNumber,
    ThermostatSwitch,
    ThermostatSwitchDescription,
    get_thermostat_parameter_types,
)

_LOGGER = logging.getLogger()


class Thermostat(VirtualDevice):
    """Represents a thermostat."""

    __slots__ = ()

    @event_listener
    async def on_event_thermostat_sensors(self, sensors: dict[str, Any]) -> bool:
        """Update thermostat sensors and dispatch the events."""
        _LOGGER.debug("Received thermostat %i sensors", self.index)
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    @event_listener
    async def on_event_thermostat_parameters(
        self, parameters: list[tuple[int, ParameterValues]]
    ) -> bool:
        """Update thermostat parameters and dispatch the events."""
        _LOGGER.debug("Received thermostat %i parameters", self.index)
        parameter_types = get_thermostat_parameter_types()

        def _thermostat_parameter_events() -> Generator[Coroutine]:
            """Get dispatch calls for thermostat parameter events."""
            for index, values in parameters:
                description = parameter_types[index]
                handler = (
                    ThermostatSwitch
                    if isinstance(description, ThermostatSwitchDescription)
                    else ThermostatNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self,
                        description=description,
                        values=values,
                        index=index,
                        offset=(self.index * len(parameters)),
                    ),
                )

        await asyncio.gather(*_thermostat_parameter_events())
        return True


__all__ = ["Thermostat"]



"""Contains tests for boiler power structure decoder."""

import pytest

from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.boiler_power import ATTR_BOILER_POWER, BoilerPowerStructure


@pytest.fixture(name="boiler_power_structure")
def fixture_boiler_power_structure() -> BoilerPowerStructure:
    """Fixture for BoilerPowerStructure."""
    return BoilerPowerStructure(frame=SensorDataMessage())


class TestBoilerPowerStructure:
    """Test the BoilerPower structure decoder."""

    def test_decode(self, boiler_power_structure: BoilerPowerStructure) -> None:
        """Test decoding of boiler power structure."""
        message = bytearray([0x00, 0x00, 0x48, 0x41])
        data, offset = boiler_power_structure.decode(message)
        assert data == {ATTR_BOILER_POWER: 12.5}
        assert offset == 4

    def test_decode_with_nan(
        self, boiler_power_structure: BoilerPowerStructure
    ) -> None:
        """Test decoding of boiler power structure with NaN value."""
        message = bytearray([0x00, 0x00, 0xC0, 0x7F])  # Represents NaN
        data, offset = boiler_power_structure.decode(message)
        assert data == {}
        assert offset == 4



"""Contains a pending alerts structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_PENDING_ALERTS: Final = "pending_alerts"


class PendingAlertsStructure(StructureDecoder):
    """Represents a pending alerts structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        alerts_number = message[offset]
        return ensure_dict(data, {ATTR_PENDING_ALERTS: alerts_number}), (
            offset + alerts_number + 1
        )


__all__ = ["ATTR_PENDING_ALERTS", "PendingAlertsStructure"]



"""Contains tests for the ecoMAX device."""

from datetime import timedelta
import logging
from typing import Any, cast
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyplumio.const import (
    ATTR_FRAME_ERRORS,
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SENSORS,
    ATTR_SETUP,
    ATTR_SIZE,
    ATTR_STATE,
    ATTR_SWITCH,
    ATTR_TYPE,
    ATTR_VALUE,
    STATE_OFF,
    STATE_ON,
    DeviceState,
    DeviceType,
    FrameType,
    State,
    UnitOfMeasurement,
)
from pyplumio.devices.ecomax import (
    ATTR_FUEL_BURNED,
    ATTR_MIXERS,
    ATTR_THERMOSTATS,
    REQUIRED,
    EcoMAX,
)
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.exceptions import RequestError
from pyplumio.frames import Request, Response
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.frames.requests import (
    AlertsRequest,
    EcomaxControlRequest,
    EcomaxParametersRequest,
    SetEcomaxParameterRequest,
    SetScheduleRequest,
    SetThermostatParameterRequest,
)
from pyplumio.frames.responses import (
    EcomaxParametersResponse,
    MixerParametersResponse,
    SchedulesResponse,
    ThermostatParametersResponse,
)
from pyplumio.parameters.ecomax import PARAMETER_TYPES, EcomaxNumber, EcomaxSwitch
from pyplumio.structures.ecomax_parameters import ATTR_ECOMAX_CONTROL
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.fuel_consumption import ATTR_FUEL_CONSUMPTION
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import (
    ATTR_MIXER_SENSORS,
    ATTR_MIXERS_AVAILABLE,
    ATTR_MIXERS_CONNECTED,
)
from pyplumio.structures.network_info import ATTR_NETWORK, NetworkInfo
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
    Schedule,
    ScheduleDay,
    ScheduleNumber,
    ScheduleSwitch,
)
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    ATTR_THERMOSTAT_PROFILE,
)
from pyplumio.structures.thermostat_sensors import (
    ATTR_THERMOSTAT_SENSORS,
    ATTR_THERMOSTATS_AVAILABLE,
    ATTR_THERMOSTATS_CONNECTED,
)
from tests.conftest import (
    UNDEFINED,
    class_from_json,
    equal_parameter_value,
    json_test_data,
)


@patch("asyncio.Queue.put_nowait")
@patch("pyplumio.devices.PhysicalDevice.handle_frame")
@patch("pyplumio.frames.Request.response", return_value=Mock(spec=Response))
async def test_ecomax_handle_frame(
    mock_response, mock_handle_frame, mock_put_nowait, ecomax: EcoMAX
) -> None:
    """Test ecoMAX frame handling."""
    request = Request()
    ecomax.handle_frame(request)
    args = mock_response.call_args[1]
    assert isinstance(args["data"][ATTR_NETWORK], NetworkInfo)
    mock_put_nowait.assert_called_once_with(mock_response.return_value)
    mock_handle_frame.assert_called_once_with(request)


@pytest.mark.parametrize(
    ("frame_type", "frame_request"),
    [
        (FrameType.REQUEST_ALERTS, AlertsRequest(recipient=DeviceType.ECOMAX)),
        (
            FrameType.REQUEST_ECOMAX_PARAMETER_CHANGES,
            EcomaxParametersRequest(recipient=DeviceType.ECOMAX),
        ),
    ],
)
@patch("asyncio.Queue.put_nowait")
@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_frame_versions_tracker(
    mock_put_nowait,
    ecomax: EcoMAX,
    sensor_data: SensorDataMessage,
    frame_type: FrameType,
    frame_request: Request,
) -> None:
    """Test frame version tracker."""

    def _frame_version_data(frame: FrameType, version: int) -> dict[str, Any]:
        """Replace frame data with new version information."""
        return {ATTR_SENSORS: {ATTR_FRAME_VERSIONS: {frame: version}}}

    # Test with frame type that are handled during setup.
    sensor_data.data = _frame_version_data(frame_type, 1)
    ecomax.data.clear()
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    mock_put_nowait.assert_not_called()

    # Test with same frame type after setup done.
    mock_put_nowait.reset_mock()
    sensor_data.data = _frame_version_data(frame_type, 2)
    ecomax.data.clear()
    ecomax.data[ATTR_SETUP] = True
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    mock_put_nowait.assert_called_once_with(frame_request)


@pytest.mark.parametrize("state", [STATE_ON, STATE_OFF])
async def test_ecomax_control(state: State, ecomax: EcoMAX, caplog) -> None:
    """Test ecoMAX control."""
    coro = getattr(ecomax, f"turn_{state}")
    await coro()
    assert "control is not available" in caplog.text
    switch = AsyncMock(spec=EcomaxSwitch)
    ecomax.data[ATTR_ECOMAX_CONTROL] = switch
    await coro()
    switch.set.assert_awaited_once_with(state)


@pytest.mark.parametrize("state", [STATE_ON, STATE_OFF])
@patch("pyplumio.devices.ecomax.EcoMAX.create_task")
def test_ecomax_control_nowait(mock_create_task, ecomax: EcoMAX, state: State) -> None:
    """Test ecoMAX control without waiting for result."""
    func = getattr(ecomax, f"turn_{state}_nowait")
    with patch(
        f"pyplumio.devices.ecomax.EcoMAX.turn_{state}", new_callable=Mock
    ) as mock_coro:
        func()

    mock_coro.assert_called_once()
    mock_create_task.assert_called_once_with(mock_coro.return_value)


@patch("pyplumio.devices.mixer.Mixer.shutdown")
@patch("pyplumio.devices.thermostat.Thermostat.shutdown")
@patch("pyplumio.devices.Device.cancel_tasks")
@json_test_data("messages/sensor_data.json", selector="message")
async def test_ecomax_shutdown(
    mock_cancel_tasks,
    mock_thermostat_shutdown,
    mock_mixer_shutdown,
    ecomax: EcoMAX,
    sensor_data_message,
) -> None:
    """Test shutting down the ecoMAX and children tasks."""
    ecomax.handle_frame(SensorDataMessage(message=sensor_data_message))
    await ecomax.wait_until_done()

    with patch("pyplumio.devices.Device.wait_until_done") as mock_wait_until_done:
        await ecomax.shutdown()

    mock_wait_until_done.assert_awaited_once()
    mock_cancel_tasks.assert_called_once()
    mock_thermostat_shutdown.assert_awaited_once()
    mock_mixer_shutdown.assert_awaited_once()


@patch("pyplumio.devices.ecomax.EcoMAX.request")
@patch("pyplumio.devices.ecomax.EcoMAX.wait_for")
async def test_ecomax_setup(mock_wait_for, mock_request, ecomax: EcoMAX) -> None:
    """Test setting up an ecoMAX entry."""
    ecomax.dispatch_nowait(ATTR_SETUP, True)
    await ecomax.wait_until_done()
    assert ATTR_FRAME_ERRORS not in ecomax.data
    assert mock_request.await_count == len(REQUIRED)
    mock_wait_for.assert_awaited_once_with(ATTR_SENSORS)


@patch(
    "pyplumio.devices.ecomax.EcoMAX.request",
    side_effect=RequestError("test", FrameType.REQUEST_ALERTS),
)
@patch("pyplumio.devices.ecomax.EcoMAX.wait_for")
async def test_ecomax_setup_errors(mock_wait_for, mock_request, ecomax: EcoMAX) -> None:
    """Test setting up an ecoMAX entry with frame errors."""
    ecomax.dispatch_nowait(ATTR_SETUP, True)
    await ecomax.wait_until_done()
    assert FrameType.REQUEST_ALERTS in ecomax.get_nowait(ATTR_FRAME_ERRORS, [])
    assert mock_request.await_count == len(REQUIRED)
    mock_wait_for.assert_awaited_once_with(ATTR_SENSORS)


@patch("asyncio.Queue.put")
@class_from_json(
    EcomaxParametersResponse,
    "responses/ecomax_parameters.json",
    arguments=("message",),
)
async def test_ecomax_parameters_event_listener(
    mock_put, ecomax: EcoMAX, ecomax_parameters: EcomaxParametersResponse
) -> None:
    """Test event listener for ecoMAX parameters."""
    ecomax.handle_frame(ecomax_parameters)
    await ecomax.wait_until_done()
    fuzzy_logic = ecomax.get_nowait("fuzzy_logic")
    assert isinstance(fuzzy_logic, EcomaxSwitch)
    assert fuzzy_logic == STATE_ON
    assert fuzzy_logic.value == STATE_ON
    fuzzy_logic_request = await fuzzy_logic.create_request()
    assert isinstance(fuzzy_logic_request, SetEcomaxParameterRequest)
    assert fuzzy_logic_request.data == {
        ATTR_INDEX: 18,
        ATTR_VALUE: 1,
    }

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(ecomax_parameters)
    await ecomax.wait_until_done()
    assert ecomax.get_nowait("fuzzy_logic") is fuzzy_logic

    # Test parameter with the step (heating_heat_curve)
    fuel_calorific_value = ecomax.get_nowait("fuel_calorific_value")
    assert isinstance(fuel_calorific_value, EcomaxNumber)
    assert equal_parameter_value(fuel_calorific_value.value, 4.6)
    assert equal_parameter_value(fuel_calorific_value.min_value, 0.1)
    assert equal_parameter_value(fuel_calorific_value.max_value, 25.0)

    # Test setting parameter with the step.
    await fuel_calorific_value.set(2.5, timeout=0)
    request = mock_put.call_args[0][0]
    assert isinstance(request, SetEcomaxParameterRequest)
    assert request.data[ATTR_VALUE] == 25

    # Test parameter with the offset (heating_heat_curve_shift)
    heating_heat_curve_shift = ecomax.get_nowait("heating_curve_shift")
    assert isinstance(heating_heat_curve_shift, EcomaxNumber)
    assert equal_parameter_value(heating_heat_curve_shift.value, 0.0)
    assert equal_parameter_value(heating_heat_curve_shift.min_value, -20.0)
    assert equal_parameter_value(heating_heat_curve_shift.max_value, 20.0)
    assert heating_heat_curve_shift.unit_of_measurement == UnitOfMeasurement.CELSIUS

    # Test setting the parameter with the offset.
    await heating_heat_curve_shift.set(1, timeout=0)
    request = mock_put.call_args[0][0]
    assert isinstance(request, SetEcomaxParameterRequest)
    assert request.data[ATTR_VALUE] == 21


@class_from_json(
    EcomaxParametersResponse,
    "unknown/unknown_ecomax_parameter.json",
    arguments=("message",),
)
async def test_unknown_ecomax_parameter(
    ecomax: EcoMAX, caplog, unknown_ecomax_parameter: EcomaxParametersResponse
) -> None:
    """Test unknown ecoMAX parameter."""
    with caplog.at_level(logging.WARNING):
        ecomax.handle_frame(unknown_ecomax_parameter)
        await ecomax.wait_until_done()

    assert "unknown ecoMAX parameter (139)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_name_collisions(ecomax: EcoMAX, sensor_data: SensorDataMessage) -> None:
    """Test that the sensor names don't collide with the parameters."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    for descriptions in PARAMETER_TYPES.values():
        collisions = [
            description.name
            for description in descriptions
            if description.name in ecomax.data
        ]
        assert not collisions, f"found collisions: {collisions}"


async def test_fuel_consumption_event_listener(
    frozen_time, caplog, ecomax: EcoMAX
) -> None:
    """Test event listener for fuel consumption."""
    frozen_time.tick(timedelta(seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 3.6}))
    await ecomax.wait_until_done()
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert equal_parameter_value(fuel_burned, 0.01)

    # Test with outdated data.
    message = "Skipping outdated fuel consumption"
    frozen_time.tick(timedelta(minutes=5, seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 1}))
    await ecomax.wait_until_done()
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert equal_parameter_value(fuel_burned, 0.01)
    assert message in caplog.text
    caplog.clear()

    frozen_time.tick(timedelta(seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 7.2}))
    await ecomax.wait_until_done()
    assert message not in caplog.text
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert equal_parameter_value(fuel_burned, 0.02)


@class_from_json(
    MixerParametersResponse,
    "responses/mixer_parameters.json",
    arguments=("message",),
)
async def test_mixer_parameters_event_listener(
    ecomax: EcoMAX, mixer_parameters: MixerParametersResponse
) -> None:
    """Test event listener for mixer parameters."""
    ecomax.handle_frame(mixer_parameters)
    await ecomax.wait_until_done()
    mixers = cast(dict[int, Mixer], ecomax.get_nowait(ATTR_MIXERS))
    assert isinstance(mixers, dict)
    assert len(mixers) == 1
    assert all(isinstance(mixer, Mixer) for mixer in mixers.values())


async def test_mixer_parameters_event_listener_without_mixers(ecomax: EcoMAX) -> None:
    """Test event listener for mixer parameters without mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_PARAMETERS: {}}))
    await ecomax.wait_until_done()
    assert not ecomax.get_nowait(ATTR_MIXER_PARAMETERS, UNDEFINED)


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_mixer_sensors_event_listener(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test event listener for mixer sensors."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    mixers = cast(dict[int, Mixer], ecomax.get_nowait(ATTR_MIXERS))
    assert isinstance(mixers, dict)
    assert len(mixers) == 1
    assert all(isinstance(mixer, Mixer) for mixer in mixers.values())
    assert ecomax.get_nowait(ATTR_MIXERS_AVAILABLE) == 5
    assert ecomax.get_nowait(ATTR_MIXERS_CONNECTED) == 1


async def test_mixer_sensors_callbacks_without_mixers(ecomax: EcoMAX) -> None:
    """Test event listener for mixer sensors without mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_SENSORS: {}}))
    await ecomax.wait_until_done()
    assert ecomax.get_nowait(ATTR_MIXER_SENSORS, UNDEFINED) is False


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_ecomax_sensors_event_listener(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test event listener for ecoMAX sensors."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    heating_target = cast(float, ecomax.get_nowait("heating_target"))
    assert equal_parameter_value(heating_target, 41.0)

    ecomax_control = ecomax.get_nowait(ATTR_ECOMAX_CONTROL)
    assert isinstance(ecomax_control, EcomaxSwitch)
    assert ecomax_control.value == STATE_OFF

    ecomax_control_request = await ecomax_control.create_request()
    assert isinstance(ecomax_control_request, EcomaxControlRequest)
    assert ecomax_control_request.data == {ATTR_VALUE: 0}

    # Check ecomax_control reference equility.
    sensor_data_message2 = SensorDataMessage(data={ATTR_STATE: DeviceState.WORKING})
    ecomax.handle_frame(sensor_data_message2)
    await ecomax.wait_until_done()
    ecomax_control2 = ecomax.get_nowait(ATTR_ECOMAX_CONTROL)
    assert isinstance(ecomax_control2, EcomaxSwitch)
    assert ecomax_control2 is ecomax_control
    assert ecomax_control2.value == STATE_ON
    ecomax_control2_request = await ecomax_control.create_request()
    assert ecomax_control2_request.data == {ATTR_VALUE: 1}


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_parameters_event_listener(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test event listener for thermostat parameters."""
    ecomax.data[ATTR_THERMOSTATS_AVAILABLE] = 3
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    thermostats = cast(dict[int, Thermostat], ecomax.get_nowait(ATTR_THERMOSTATS))
    assert isinstance(thermostats, dict)
    assert len(thermostats) == 1
    assert all(
        isinstance(thermostat, Thermostat) for thermostat in thermostats.values()
    )


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_parameters_event_listener_without_thermostats(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test event listener for thermostat parameters without any thermostats."""
    ecomax.data[ATTR_THERMOSTATS_AVAILABLE] = 0
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    assert ecomax.get_nowait(ATTR_THERMOSTAT_PARAMETERS, UNDEFINED) is False


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_profile_event_listener(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test event listener for thermostat profile."""
    ecomax.data[ATTR_THERMOSTATS_AVAILABLE] = 3
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    thermostat_profile = ecomax.get_nowait(ATTR_THERMOSTAT_PROFILE)
    assert isinstance(thermostat_profile, EcomaxNumber)
    assert equal_parameter_value(thermostat_profile.value, 0.0)
    assert equal_parameter_value(thermostat_profile.min_value, 0.0)
    assert equal_parameter_value(thermostat_profile.max_value, 5.0)
    thermostat_profile_request = await thermostat_profile.create_request()
    assert isinstance(thermostat_profile_request, SetThermostatParameterRequest)
    assert thermostat_profile_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 0,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }

    # Test when thermostat profile is none.
    ecomax.handle_frame(Response(data={ATTR_THERMOSTAT_PROFILE: None}))
    await ecomax.wait_until_done()
    assert ecomax.get_nowait(ATTR_THERMOSTAT_PROFILE, UNDEFINED) is None


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_thermostat_sensors_event_listener(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test event listener for thermostat sensors."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    thermostats = cast(dict[int, Thermostat], ecomax.get_nowait(ATTR_THERMOSTATS))
    assert isinstance(thermostats, dict)
    assert len(thermostats) == 1
    assert all(
        isinstance(thermostat, Thermostat) for thermostat in thermostats.values()
    )
    assert ecomax.get_nowait(ATTR_THERMOSTATS_AVAILABLE) == 2
    assert ecomax.get_nowait(ATTR_THERMOSTATS_CONNECTED) == 1


async def test_thermostat_sensors_event_listener_without_thermostats(
    ecomax: EcoMAX,
) -> None:
    """Test event lister for thermostat sensors without thermostats."""
    ecomax.handle_frame(
        ThermostatParametersResponse(data={ATTR_THERMOSTAT_SENSORS: {}})
    )
    await ecomax.wait_until_done()
    assert ecomax.get_nowait(ATTR_THERMOSTAT_SENSORS, UNDEFINED) is False


@class_from_json(SchedulesResponse, "responses/schedules.json", arguments=("message",))
@json_test_data("responses/schedules.json", selector="data")
async def test_ecomax_schedules_event_listener(
    ecomax: EcoMAX, schedules: SchedulesResponse, schedules_data
) -> None:
    """Test event listener for schedules."""
    ecomax.handle_frame(schedules)
    await ecomax.wait_until_done()
    schedules_arg = cast(dict[str, Schedule], ecomax.get_nowait(ATTR_SCHEDULES))
    assert isinstance(schedules_arg, dict)
    assert len(schedules_arg) == 2
    heating_schedule = schedules_arg["heating"]
    assert isinstance(heating_schedule, Schedule)

    heating_schedule_switch = ecomax.get_nowait("heating_schedule_switch")
    assert isinstance(heating_schedule_switch, ScheduleSwitch)
    assert heating_schedule_switch.value == STATE_OFF

    water_heater_schedule_parameter = ecomax.get_nowait(
        "water_heater_schedule_parameter"
    )
    assert isinstance(water_heater_schedule_parameter, ScheduleNumber)
    assert water_heater_schedule_parameter.value == 5
    assert water_heater_schedule_parameter.min_value == 0
    assert water_heater_schedule_parameter.max_value == 30
    water_heater_schedule_parameter_request = (
        await water_heater_schedule_parameter.create_request()
    )
    assert isinstance(water_heater_schedule_parameter_request, SetScheduleRequest)
    assert water_heater_schedule_parameter_request.data == {
        ATTR_TYPE: "water_heater",
        ATTR_SWITCH: ecomax.data[f"water_heater_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: ecomax.data[f"water_heater_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: ecomax.data[ATTR_SCHEDULES]["water_heater"],
    }

    schedule_data = schedules_data[ATTR_SCHEDULES][0][1]
    for index, weekday in enumerate(
        (
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        )
    ):
        schedule = getattr(heating_schedule, weekday)
        assert (
            schedule.schedule
            == ScheduleDay.from_iterable(schedule_data[index]).schedule
        )

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(schedules)
    await ecomax.wait_until_done()
    assert ecomax.get_nowait("heating_schedule_switch") is heating_schedule_switch



"""Contains tests for the thermostat virtual device."""

import asyncio
from typing import Final

import pytest

from pyplumio.const import ATTR_SENSORS
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.devices.thermostat import Thermostat
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.frames.responses import ThermostatParametersResponse
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.thermostat import ThermostatNumber, ThermostatParameter
from pyplumio.structures.thermostat_parameters import ATTR_THERMOSTAT_PARAMETERS
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTAT_SENSORS
from tests.conftest import class_from_json

THERMOSTAT_INDEX: Final = 0


@pytest.fixture(name="thermostat")
def fixture_thermostat(ecomax: EcoMAX) -> Thermostat:
    """Return a Thermostat object."""
    return Thermostat(asyncio.Queue(), parent=ecomax, index=THERMOSTAT_INDEX)


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("data",))
async def test_thermostat_sensors_event_listener(
    thermostat: Thermostat, sensor_data: SensorDataMessage
) -> None:
    """Test event listener for thermostat sensors."""
    thermostat_sensors_data = sensor_data.data[ATTR_SENSORS][ATTR_THERMOSTAT_SENSORS]
    thermostat.dispatch_nowait(
        ATTR_THERMOSTAT_SENSORS, thermostat_sensors_data[THERMOSTAT_INDEX]
    )
    await thermostat.wait_until_done()
    assert thermostat.data == {
        "state": 3,
        "current_temp": 43.5,
        "target_temp": 50.0,
        "contacts": True,
        "schedule": False,
        "thermostat_sensors": True,
    }


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("data",),
)
@pytest.mark.parametrize(
    ("name", "cls", "values"),
    [
        (
            "mode",
            ThermostatNumber,
            ParameterValues(value=0, min_value=0, max_value=7),
        ),
        (
            "party_target_temp",
            ThermostatNumber,
            ParameterValues(value=220, min_value=100, max_value=350),
        ),
        (
            "holidays_target_temp",
            ThermostatNumber,
            ParameterValues(value=150, min_value=100, max_value=350),
        ),
        (
            "correction",
            ThermostatNumber,
            ParameterValues(value=100, min_value=60, max_value=140),
        ),
        (
            "away_timer",
            ThermostatNumber,
            ParameterValues(value=2, min_value=0, max_value=60),
        ),
        (
            "airing_timer",
            ThermostatNumber,
            ParameterValues(value=1, min_value=0, max_value=60),
        ),
        (
            "party_timer",
            ThermostatNumber,
            ParameterValues(value=1, min_value=0, max_value=60),
        ),
        (
            "holidays_timer",
            ThermostatNumber,
            ParameterValues(value=10, min_value=0, max_value=60),
        ),
        (
            "hysteresis",
            ThermostatNumber,
            ParameterValues(value=9, min_value=0, max_value=50),
        ),
        (
            "day_target_temp",
            ThermostatNumber,
            ParameterValues(value=222, min_value=100, max_value=350),
        ),
        (
            "night_target_temp",
            ThermostatNumber,
            ParameterValues(value=212, min_value=100, max_value=350),
        ),
        (
            "antifreeze_target_temp",
            ThermostatNumber,
            ParameterValues(value=90, min_value=50, max_value=300),
        ),
    ],
)
async def test_thermostat_parameters_event_listener(
    thermostat: Thermostat,
    thermostat_parameters: ThermostatParametersResponse,
    name: str,
    cls: type[ThermostatParameter],
    values: ParameterValues,
) -> None:
    """Test event listener for thermostat parameters."""
    thermostat_parameters_data = thermostat_parameters.data[ATTR_THERMOSTAT_PARAMETERS]
    thermostat.dispatch_nowait(
        ATTR_THERMOSTAT_PARAMETERS, thermostat_parameters_data[THERMOSTAT_INDEX]
    )
    await thermostat.wait_until_done()
    assert len(thermostat.data) == 13
    parameter = thermostat.get_nowait(name)
    assert isinstance(parameter, cls)
    assert parameter.values == values



"""Contains tests for the parameter descriptors."""

from __future__ import annotations

import asyncio
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyplumio.const import BYTE_UNDEFINED, STATE_OFF, STATE_ON, State, UnitOfMeasurement
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.frames import Request
from pyplumio.parameters import (
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    is_valid_parameter,
)


@pytest.fixture(name="number")
def fixture_number(ecomax: EcoMAX) -> Number:
    """Numeric parameter instance.

    Returns a Number object for testing.
    """
    return Number(
        device=ecomax,
        values=ParameterValues(value=1, min_value=0, max_value=5),
        description=NumberDescription(
            name="test_number", unit_of_measurement=UnitOfMeasurement.CELSIUS
        ),
    )


@pytest.fixture(name="switch")
def fixture_switch(ecomax: EcoMAX) -> Switch:
    """Switch instance.

    Returns a Switch object for testing.
    """
    return Switch(
        device=ecomax,
        values=ParameterValues(value=0, min_value=0, max_value=1),
        description=SwitchDescription(name="test_switch"),
    )


@pytest.mark.parametrize(
    ("data", "expected_result"),
    [
        (bytearray([BYTE_UNDEFINED, 0xFE, BYTE_UNDEFINED, BYTE_UNDEFINED]), True),
        (
            bytearray([BYTE_UNDEFINED, BYTE_UNDEFINED, BYTE_UNDEFINED, BYTE_UNDEFINED]),
            False,
        ),
    ],
)
def test_is_valid_parameter(data: bytearray, expected_result: bool) -> None:
    """Test is_valid_parameter.

    Checks if parameter data is valid.
    """
    assert is_valid_parameter(data) is expected_result


class DummyParameter(Parameter):
    """Dummy parameter for testing."""

    def _pack_value(self, value: Any) -> int:
        """Pack the parameter value."""
        return int(value)

    def _unpack_value(self, value: Any):
        """Unpack the parameter value."""

    def validate(self, value: Any):
        """Validate a parameter value."""

    @property
    def value(self) -> Any:
        """Return the value."""

    @property
    def min_value(self) -> Any:
        """Return the minimum allowed value."""

    @property
    def max_value(self) -> Any:
        """Return the maximum allowed value."""

    @property
    async def create_request(self):
        """Create a request to change the parameter."""


@pytest.fixture(name="parameter")
def fixture_parameter(ecomax: EcoMAX) -> Parameter:
    """Parameter instance.

    Returns a DummyParameter object for testing.
    """
    return DummyParameter(
        device=ecomax,
        values=ParameterValues(value=6, min_value=0, max_value=10),
        description=ParameterDescription(name="test_param"),
    )


class TestParameter:
    """Tests for Parameter class.

    Verifies representation, math, equality, copy, set, and update.
    """

    def test_repr(self, parameter: Parameter) -> None:
        """Test __repr__.

        Checks parameter representation.
        """
        assert repr(parameter) == (
            f"DummyParameter(device={parameter.device}, "
            "description=ParameterDescription(name='test_param', optimistic=False), "
            "values=ParameterValues(value=6, min_value=0, max_value=10), "
            "index=0)"
        )

    def test_int(self, parameter: Parameter) -> None:
        """Test __int__.

        Checks parameter conversion to int.
        """
        result = int(parameter)
        assert result == 6
        assert isinstance(result, int)

    def test_hash(self, parameter: Parameter) -> None:
        """Test __hash__.

        Checks parameter hashing.
        """
        assert hash(parameter) == hash(
            frozenset({("value", 6), ("min_value", 0), ("max_value", 10)})
        )

    @pytest.mark.parametrize(
        ("func_name", "other", "expected_result", "expected_type"),
        [
            ("__add__", "Parameter", 12, int),
            ("__add__", ParameterValues(1, 0, 10), 7, int),
            ("__add__", 10, 16, int),
            ("__add__", 2.1, 8, int),
            ("__add__", "mango", NotImplemented, None),
            ("__sub__", "Parameter", 0, int),
            ("__sub__", ParameterValues(3, 0, 10), 3, int),
            ("__sub__", 4, 2, int),
            ("__sub__", 4.1, 2, int),
            ("__sub__", None, NotImplemented, None),
            ("__truediv__", "Parameter", 1, float),
            ("__truediv__", ParameterValues(3, 0, 10), 2, float),
            ("__truediv__", 3, 2, float),
            ("__truediv__", 2.2, 3, float),
            ("__floordiv__", "Parameter", 1, int),
            ("__floordiv__", ParameterValues(2, 0, 10), 3, int),
            ("__floordiv__", 4, 1, int),
            ("__floordiv__", 1.5, 6, int),
            ("__floordiv__", True, 6, None),
            ("__mul__", "Parameter", 36, int),
            ("__mul__", ParameterValues(2, 0, 10), 12, int),
            ("__mul__", 3, 18, int),
            ("__mul__", 2.5, 12, int),
            ("__mul__", object, NotImplemented, None),
        ],
    )
    def test_math(
        self,
        func_name: str,
        other: Any,
        expected_result: int,
        expected_type: type,
        parameter: Parameter,
    ) -> None:
        """Test math operations.

        Checks addition, subtraction, division, multiplication, etc.
        """
        if other == "Parameter":
            other = parameter

        func = getattr(parameter, func_name)
        result = func(other)
        assert result == expected_result
        if expected_type:
            assert isinstance(result, expected_type)

    @pytest.mark.parametrize(
        ("func_name", "other", "expected_result"),
        [
            ("__eq__", "Parameter", True),
            ("__eq__", ParameterValues(7, 1, 10), False),
            ("__eq__", ParameterValues(6, 2, 10), True),
            ("__eq__", 6, True),
            ("__eq__", 2.5, False),
            ("__eq__", True, False),
            ("__eq__", "banana", NotImplemented),
            ("__ge__", "Parameter", True),
            ("__ge__", ParameterValues(6, 1, 10), True),
            ("__ge__", ParameterValues(9, 2, 10), False),
            ("__ge__", 6, True),
            ("__ge__", 2.5, True),
            ("__ge__", True, True),
            ("__ge__", None, NotImplemented),
            ("__gt__", "Parameter", False),
            ("__gt__", ParameterValues(4, 1, 10), True),
            ("__gt__", ParameterValues(6, 2, 10), False),
            ("__gt__", 7, False),
            ("__gt__", 2.5, True),
            ("__gt__", True, True),
            ("__gt__", bytearray(), NotImplemented),
            ("__le__", "Parameter", True),
            ("__le__", ParameterValues(6, 1, 10), True),
            ("__le__", ParameterValues(4, 2, 10), False),
            ("__le__", 8, True),
            ("__le__", 2.5, False),
            ("__le__", True, False),
            ("__le__", b"", NotImplemented),
            ("__lt__", "Parameter", False),
            ("__lt__", ParameterValues(7, 1, 10), True),
            ("__lt__", ParameterValues(4, 2, 10), False),
            ("__lt__", 7, True),
            ("__lt__", 2.5, False),
            ("__lt__", True, False),
            ("__lt__", object, NotImplemented),
        ],
    )
    def test_eq(
        self,
        func_name: str,
        other: Any,
        expected_result: bool,
        parameter: Parameter,
    ) -> None:
        """Test equality and comparison.

        Checks eq, ge, gt, le, lt.
        """
        if other == "Parameter":
            other = parameter

        func = getattr(parameter, func_name)
        assert func(other) is expected_result

    def test_copy(self, parameter: Parameter) -> None:
        """Test __copy__.

        Checks shallow copying of parameter.
        """
        parameter_copy = parameter.__copy__()
        assert parameter_copy.values is not parameter.values

    @patch.object(DummyParameter, "validate")
    @patch.object(DummyParameter, "create_request", new_callable=AsyncMock)
    @patch("asyncio.Queue.put")
    @pytest.mark.parametrize("nowait", [False, True])
    async def test_set(
        self,
        mock_put,
        mock_create_request,
        mock_validate,
        parameter: Parameter,
        nowait: bool,
    ) -> None:
        """Test set.

        Checks setting a parameter with and without waiting.
        """
        if nowait:
            parameter.set_nowait(4)
            await parameter.device.wait_until_done()
        else:
            mock_update_pending = AsyncMock(spec=asyncio.Event)
            mock_update_done = AsyncMock(spec=asyncio.Event)
            mock_update_done.is_set = Mock(return_value=True)
            with (
                patch.object(DummyParameter, "update_pending", mock_update_pending),
                patch.object(DummyParameter, "update_done", mock_update_done),
            ):
                result = await parameter.set(4)

            assert result is True
            mock_update_done.clear.assert_called_once()
            mock_update_pending.set.assert_called_once()

        assert parameter == 4
        mock_validate.assert_called_once_with(4)
        mock_create_request.assert_awaited_once()
        mock_put.assert_awaited_once_with(mock_create_request.return_value)

    @patch.object(DummyParameter, "validate")
    @patch.object(DummyParameter, "create_request", new_callable=AsyncMock)
    @patch("asyncio.Queue.put")
    @pytest.mark.parametrize("nowait", [False, True])
    async def test_set_unchanged(
        self,
        mock_put,
        mock_create_request,
        mock_validate,
        parameter: Parameter,
        nowait: bool,
    ) -> None:
        """Test set with unchanged value.

        Checks setting a parameter when value is unchanged.
        """
        if nowait:
            parameter.set_nowait(6)
            await parameter.device.wait_until_done()
        else:
            result = await parameter.set(6)
            assert result is True

        assert parameter == 6
        mock_validate.assert_called_once_with(6)
        mock_create_request.assert_not_awaited()
        mock_put.assert_not_awaited()

    @patch.object(DummyParameter, "validate")
    @patch.object(DummyParameter, "create_request", new_callable=AsyncMock)
    @patch("asyncio.Queue.put")
    @pytest.mark.parametrize(
        (
            "event_side_effect",
            "retries",
            "expected_result",
            "expected_call_count",
            "log_message",
        ),
        [
            ((False, True), 5, True, 2, None),
            (
                (False, False),
                2,
                False,
                2,
                "Unable to confirm update of 'test_param' parameter after 2 retries",
            ),
        ],
    )
    async def test_set_with_retries(
        self,
        mock_put,
        mock_create_request,
        mock_validate,
        parameter: Parameter,
        caplog,
        event_side_effect: tuple[bool, ...],
        retries: int,
        expected_result: bool,
        expected_call_count: int,
        log_message: str | None,
    ) -> None:
        """Test set without retries.

        Checks setting a parameter with and without retries.
        """
        parameter.description = ParameterDescription(name="test_param")
        mock_update_pending = AsyncMock(spec=asyncio.Event)
        mock_update_done = AsyncMock(spec=asyncio.Event)
        mock_update_done.is_set = Mock(side_effect=event_side_effect)

        with (
            patch.object(DummyParameter, "update_pending", mock_update_pending),
            patch.object(DummyParameter, "update_done", mock_update_done),
        ):
            result = await parameter.set(5, retries=retries)

        assert result is expected_result
        assert parameter == 5
        mock_create_request.assert_awaited_once()
        mock_validate.assert_called_once_with(5)
        assert mock_put.call_count == expected_call_count

        if log_message:
            assert log_message in caplog.text

    @patch.object(DummyParameter, "validate")
    @patch.object(DummyParameter, "create_request", new_callable=AsyncMock)
    @patch("asyncio.Queue.put")
    @patch("asyncio.Event.set")
    async def test_set_optimistic(
        self,
        mock_set,
        mock_put,
        mock_create_request,
        mock_validate,
        parameter: Parameter,
    ) -> None:
        """Test set without retries.

        Checks setting a parameter with and without retries.
        """
        parameter.description = ParameterDescription(name="test_param", optimistic=True)
        result = await parameter.set(5, retries=3)
        assert result is True
        assert parameter == 5
        assert parameter.update_pending.is_set() is False
        mock_validate.assert_called_once_with(5)
        mock_put.assert_awaited_once_with(mock_create_request.return_value)
        mock_set.assert_not_called()

    @pytest.mark.usefixtures("skip_asyncio_events")
    @patch.object(DummyParameter, "create_request", new_callable=AsyncMock)
    async def test_update(self, mock_create_request, parameter: Parameter) -> None:
        """Test update.

        Checks updating a parameter.
        """
        await parameter.set(5)
        parameter_values = ParameterValues(value=5, min_value=1, max_value=10)
        parameter.update(parameter_values)
        assert parameter.values == parameter_values
        mock_create_request.assert_awaited_once()

    def test_create_or_update_parameter(self, ecomax: EcoMAX) -> None:
        """Test create_or_update.

        Checks creating or updating a parameter.
        """
        description = ParameterDescription(name="test")
        values = ParameterValues(value=3, min_value=0, max_value=5)
        with patch("pyplumio.parameters.Parameter.update") as mock_update:
            parameter = DummyParameter.create_or_update(
                device=ecomax, description=description, values=values
            )

        mock_update.assert_not_called()
        assert isinstance(parameter, DummyParameter)

        # Test updating an existing parameter.
        ecomax.data[description.name] = parameter
        with patch("pyplumio.parameters.Parameter.update") as mock_update:
            DummyParameter.create_or_update(
                device=ecomax, description=description, values=values
            )

        mock_update.assert_called_once()


class TestNumber:
    """Tests for Number class.

    Verifies value, validation, set, and request creation.
    """

    def test_values(self, number: Number) -> None:
        """Test values.

        Checks number value, min, and max.
        """
        assert number.value == 1
        assert number.min_value == 0
        assert number.max_value == 5

    def test_validate(self, number: Number) -> None:
        """Test validate.

        Checks number validation.
        """
        assert number.validate(2) is True

        with pytest.raises(ValueError, match="The value must be between"):
            number.validate(6)

        with pytest.raises(ValueError, match="value must be adjusted in increments"):
            number.validate(2.1)

    @patch("pyplumio.parameters.Parameter.set")
    async def test_set(self, mock_set, number: Number) -> None:
        """Test set.

        Checks setting a number.
        """
        await number.set(5)
        mock_set.assert_awaited_once_with(5, retries=0, timeout=5.0)

    @patch("pyplumio.parameters.Parameter.set_nowait")
    def test_set_nowait(self, mock_set_nowait, number: Number) -> None:
        """Test set_nowait.

        Checks setting a number without waiting.
        """
        number.set_nowait(5)
        mock_set_nowait.assert_called_once_with(5, retries=0, timeout=5.0)

    async def test_create_request(self, number: Number) -> None:
        """Test create_request.

        Checks creating a number change request.
        """
        assert isinstance(await number.create_request(), Request)


class TestSwitch:
    """Tests for Switch class.

    Verifies value, validation, set, and state changes.
    """

    def test_value(self, switch: Switch) -> None:
        """Test value.

        Checks switch value, min, and max.
        """
        assert switch.value == STATE_OFF
        assert switch.min_value == STATE_OFF
        assert switch.max_value == STATE_ON

    def test_validate(self, switch: Switch) -> None:
        """Test validate.

        Checks switch validation.
        """
        assert switch.validate(STATE_ON) is True

        with pytest.raises(ValueError, match="must be either 'on', 'off' or boolean"):
            switch.validate(2)

    @patch("pyplumio.parameters.Parameter.set")
    @pytest.mark.parametrize("state", [True, False, STATE_ON, STATE_OFF])
    async def test_set(self, mock_set, switch: Switch, state: State | bool) -> None:
        """Test set.

        Checks setting a switch.
        """
        await switch.set(state)
        mock_set.assert_awaited_once_with(state, retries=0, timeout=5.0)

    @patch("pyplumio.parameters.Parameter.set_nowait")
    @pytest.mark.parametrize("state", [True, False, STATE_ON, STATE_OFF])
    def test_set_nowait(
        self, mock_set_nowait, switch: Switch, state: State | bool
    ) -> None:
        """Test set_nowait.

        Checks setting a switch without waiting.
        """
        switch.set_nowait(state)
        mock_set_nowait.assert_called_once_with(state, retries=0, timeout=5.0)

    async def test_create_request(self, switch: Switch) -> None:
        """Test create_request.

        Checks creating a switch change request.
        """
        assert isinstance(await switch.create_request(), Request)

    @patch("pyplumio.parameters.Switch.set")
    async def test_turn_on(self, mock_set, switch: Switch) -> None:
        """Test turn_on.

        Checks turning switch on.
        """
        await switch.turn_on()
        mock_set.assert_called_once_with(STATE_ON)

    @patch("pyplumio.parameters.Switch.set")
    async def test_turn_off(self, mock_set, switch: Switch) -> None:
        """Test turn_off.

        Checks turning switch off.
        """
        await switch.turn_off()
        mock_set.assert_called_once_with(STATE_OFF)

    @patch("pyplumio.parameters.Switch.set_nowait")
    async def test_switch_turn_on_nowait(self, mock_set_nowait, switch: Switch) -> None:
        """Test turn_on_nowait.

        Checks turning switch on without waiting.
        """
        switch.turn_on_nowait()
        mock_set_nowait.assert_called_once_with(STATE_ON)

    @patch("pyplumio.parameters.Switch.set_nowait")
    async def test_switch_turn_off_nowait(
        self, mock_set_nowait, switch: Switch
    ) -> None:
        """Test turn_off_nowait.

        Checks turning switch off without waiting.
        """
        switch.turn_off_nowait()
        mock_set_nowait.assert_called_once_with(STATE_OFF)

    @pytest.mark.parametrize(
        ("state", "expected_result"),
        [(STATE_ON, True), (STATE_OFF, False), (True, True), (False, False)],
    )
    def test_states(
        self,
        state: State | bool,
        expected_result: bool,
        switch: Switch,
    ) -> None:
        """Test states.

        Checks switch equality using integers and states.
        """
        switch.update(ParameterValues(1, 0, 1))
        assert switch.__eq__(state) is expected_result



"""Contains a tests for the message frame classes."""

from typing import Final

import pytest
from tests.conftest import json_test_data, load_json_parameters, load_json_test_data

from pyplumio.const import ATTR_SENSORS, ATTR_STATE
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.frames.messages import RegulatorDataMessage, SensorDataMessage
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.regulator_data import ATTR_REGDATA


@pytest.mark.parametrize(
    ("schema", "regdata"),
    zip(
        load_json_test_data("responses/regulator_data_schema.json"),
        load_json_test_data("messages/regulator_data.json"),
    ),
    ids=[
        "unknown_regulator_data_version",
        "EM350P2_regulator_data",
        "incomplete_boolean",
    ],
)
async def test_regulator_data_message(ecomax: EcoMAX, schema, regdata) -> None:
    """Test a regulator data message."""
    frame = RegulatorDataMessage(message=regdata["message"])
    frame.assign_to(ecomax)
    ecomax.load_nowait(schema["data"])
    await ecomax.wait_until_done()

    if regdata["id"] == "unknown_regulator_data_version":
        assert ATTR_FRAME_VERSIONS not in frame.data
        assert ATTR_REGDATA not in frame.data
    else:
        assert frame.data[ATTR_FRAME_VERSIONS] == regdata["data"][ATTR_FRAME_VERSIONS]
        assert frame.data[ATTR_REGDATA] == regdata["data"][ATTR_REGDATA]


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("messages/sensor_data.json"),
)
def test_sensor_data_message(message, data) -> None:
    """Test a sensor data message."""
    assert SensorDataMessage(message=message).data == data


INDEX_STATE: Final = 22


@json_test_data("messages/sensor_data.json", selector="message")
async def test_sensor_data_message_with_unknown_state(sensor_data_message) -> None:
    """Test a sensor data message with an unknown device state."""
    sensor_data_message[INDEX_STATE] = 99
    sensor_data = SensorDataMessage(message=sensor_data_message)
    assert sensor_data.data[ATTR_SENSORS][ATTR_STATE] == 99
